// ***************************************************************************************************
// ***************************************************************************************************
//  RELIGHTING FRAMEWORK v0.10
//  Sebastian Enrique - senrique@cs.columbia.edu
//  Columbia University
//  January 2004
// ***************************************************************************************************
//  TGAImage.cpp implementation of TGAImage class to load and save TGA files
// ***************************************************************************************************

// TGAImage.cpp: implementation of the TGAimage class.
//
//////////////////////////////////////////////////////////////////////

#include "TGAImage.h"
#include "stdio.h"


// =================================================================================================================
//  save
//  saves a TGA file with current data
// -----------------------------------------------------------------------------------------------------------------
void TGAImage::save(char* filename) {
	FILE * fp = fopen(filename, "wb");
	if (fp != NULL) {
		byte header[18];
		header[0] = 0;		// size of ID field that follows 18 byte header (0 usually)
		header[1] = 0;		// type of colour map 0=none, 1=has palette
		header[2] = 2;		// type of image 0=none,1=indexed,2=rgb,3=grey,+8=rle packed
		header[3] = 0;		// first colour map entry in palette
		header[4] = 0;
		header[5] = 0;		// number of colours in palette
		header[6] = 0;
		header[7] = 0;		// number of bits per palette entry 15,16,24,32

		header[8] = 0;		// image x origin
		header[9] = 0;
		header[10] = 0;		// image y origin
		header[11] = 0;
		header[12] = (byte)((width & 0xFF) >> 0);	// image width in pixels low order bits
		header[13] = (byte)((width & 0xFF00) >> 8);	// image width in pixels high order bits
		header[14] = (byte)((height & 0xFF) >> 0);	// image height in pixels low order bits
		header[15] = (byte)((height & 0xFF00) >> 8);	// image height in pixels high order bits
		header[16] = 24;	// image bits per pixel 8,16,24,32
		header[17] = 8;		// image descriptor bits (vh flip bits) I just copied '8' from Photoshop
		//switch R and B from RGB to BGR
		//the coming up loop may also require inverting the image on y
		int totallength = width*height*3;
		unsigned char * bgraData;
		bgraData = new unsigned char[totallength];
		for (int i = 0; i < totallength; i += 3) {
			bgraData[i+2] = data[i];
			bgraData[i+1] = data[i+1];
			bgraData[i+0] = data[i+2];
		}
		fwrite(header, sizeof(byte), 18, fp);
		fwrite(bgraData, sizeof(byte), totallength, fp);
		fclose(fp);
		delete bgraData;
	}
}
// =================================================================================================================


// =================================================================================================================
//  open
//  opens a TGA file
// -----------------------------------------------------------------------------------------------------------------
//fill the image data from the file filename
//only works on 32 bit TGA images
void TGAImage::open(char* filename)
{
	int totalLength;
	FILE *iFile;
	if (iFile = fopen(filename, "rb"))
	{
		byte header[18];
		fread(&header, sizeof(byte), 18, iFile);
		width = (header[13] << 8) + header[12];
		height = (header[15] << 8) + header[14];
		totalLength = width*height*3;
		data = new unsigned char[width*height*3]; //always stored internally as RGB
		fseek(iFile, header[0], SEEK_CUR);
		fread(data, sizeof(byte), totalLength, iFile);
		fclose(iFile);
		unsigned char temp;
		//switch R and B from BGR to RGB
		for (int i = 0; i < totalLength; i += 3) {
			temp = data[i];
			data[i] = data[i + 2];
			data[i + 2] = temp;
		}
	}
}
// =================================================================================================================


// =================================================================================================================
//  getFace
//  gets a face from a cross like TGA file cubemap
// -----------------------------------------------------------------------------------------------------------------
void TGAImage::getFace(int face, unsigned char *facedata)
{
	long facewidth, faceheight;
	long i, j;

	facewidth = (width / 3) * 3;
	faceheight = (height / 4);

	switch(face)
	{
	case 0:	// +X
		for(i=0; i<faceheight; i++)
			for(j=0; j<facewidth; j++)
				facedata[i*facewidth + j] = data[(2*facewidth) + (2*3*facewidth*faceheight) + i*width*3 + j];
		this->flipY(facedata, (width / 3), faceheight);
		break;
	case 1:	// -X
		for(i=0; i<faceheight; i++)
			for(j=0; j<facewidth; j++)
				facedata[i*facewidth + j] = data[(2*3*facewidth*faceheight) + i*width*3 + j];
		this->flipY(facedata, (width / 3), faceheight);
		break;
	case 5:	// -Z
		for(i=0; i<faceheight; i++)
			for(j=0; j<facewidth; j++)
				facedata[i*facewidth + j] = data[(facewidth) + i*width*3 + j];
		this->flipX(facedata, (width / 3), faceheight);
		break;
	case 4:	// +Z
		for(i=0; i<faceheight; i++)
			for(j=0; j<facewidth; j++)
				facedata[i*facewidth + j] = data[(facewidth) + (3*2*facewidth*faceheight) + i*width*3 + j];
		this->flipY(facedata, (width / 3), faceheight);
		break;
	case 2:	// +Y
		for(i=0; i<faceheight; i++)
			for(j=0; j<facewidth; j++)
				facedata[i*facewidth + j] = data[(facewidth) + (3*3*facewidth*faceheight) + i*width*3 + j];
		this->flipY(facedata, (width / 3), faceheight);
		break;
	case 3:	// -Y
		for(i=0; i<faceheight; i++)
			for(j=0; j<facewidth; j++)
				facedata[i*facewidth + j] = data[(facewidth) + (3*facewidth*faceheight) + i*width*3 + j];
		this->flipY(facedata, (width / 3), faceheight);
		break;
	case 6:	// 
		for(i=0; i<faceheight; i++)
			for(j=0; j<facewidth; j++)
				facedata[i*facewidth + j] = 0;
		break;
	}	
}
// =================================================================================================================


// =================================================================================================================
//  flipY
//  flips an image in Y axis
// -----------------------------------------------------------------------------------------------------------------
void TGAImage::flipY(unsigned char *data_, int scanline_width, int num_scanlines)
{
	unsigned char buffer;
	int i, j;

	for(j=0;j<(num_scanlines/2);j++)
	{
		for(i=0;i<(scanline_width*3);i++)
		{
			buffer = data_[3*j*scanline_width+i];
			data_[3*j*scanline_width+i] = data_[3*(num_scanlines - j - 1)*scanline_width+i];
			data_[3*(num_scanlines - j - 1)*scanline_width+i] = buffer;
		}
	}
}
// =================================================================================================================


// =================================================================================================================
//  flipX
//  flips an image in X axis
// -----------------------------------------------------------------------------------------------------------------
void TGAImage::flipX(unsigned char *data_, int scanline_width, int num_scanlines)
{
	unsigned char buffer;
	int i, j;

	for(j=0;j<(num_scanlines);j++)
	{
		for(i=0;i<(scanline_width/2);i++)
		{
			buffer = data_[3*j*scanline_width+3*i];
			data_[3*j*scanline_width+3*i] = data_[3*j*scanline_width+(3*(scanline_width-1)-3*i)];
			data_[3*j*scanline_width+(3*(scanline_width-1)-3*i)] = buffer;

			buffer = data_[3*j*scanline_width+3*i+1];
			data_[3*j*scanline_width+3*i+1] = data_[3*j*scanline_width+(3*(scanline_width-1)-3*i)+1];
			data_[3*j*scanline_width+(3*(scanline_width-1)-3*i)+1] = buffer;

			buffer = data_[3*j*scanline_width+3*i+2];
			data_[3*j*scanline_width+3*i+2] = data_[3*j*scanline_width+(3*(scanline_width-1)-3*i)+2];
			data_[3*j*scanline_width+(3*(scanline_width-1)-3*i)+2] = buffer;
		}
	}
}
// =================================================================================================================

// ***************************************************************************************************
// ***************************************************************************************************
