// ***************************************************************************************************
// ***************************************************************************************************
//  RELIGHTING FRAMEWORK v0.10
//  Sebastian Enrique - senrique@cs.columbia.edu
//  Columbia University
//  January 2004
// ***************************************************************************************************
//  cmcalc.cpp contains the functions used for some cubemap computations like generation of
//  distant light directions, copy of them, multiplication of them by a rotation matrix, and
//  multiplation by rotation matrix to get corresponding texels
// ***************************************************************************************************

#include "cmcalc.h"
#include <math.h>
#include <windows.h>											// Header File For Windows

// EXPECTED GLOBAL VARIABLES
extern unsigned int RF_LIGHTRES;	// global variable containing size of the cubemap grid: 6 * RF_LIGHTRES * RF_LIGHTRES
extern unsigned long RF_IMAGES;		// global variable, RF_IMAGES =  6 * RF_LIGHTRES * RF_LIGHTRES
extern unsigned long RF_BLOCKS;		// global variable, the image is divided in RF_BLOCKS = RF_BLOCKRESH * RF_BLOCKRESW blocks
extern double *weights;				// global variable, a vector holding weights of each light contribution -size RF_IMAGES-
extern double (*s)[3];				// global variable, a vector holding each light contribution -size RF_IMAGES-
extern double exposure;				// global variable, a factor to linearly increase or decrease the whole environment map contribution


// =================================================================================================================
//  generateCubeMapDirections
//  creates in vector cmdirs the direction of each one of the texels that will be used for relighting computation,
//  generates the corresponding weighting factor as well.
//  directions are normalized or not depending on the parameter.
// -----------------------------------------------------------------------------------------------------------------
void generateCubeMapDirections(float *cmdirs, bool normalize, double *weightsvector)
{
	int i;
	int x, y;
	float tc, sc;
	float v[3];
	float len, invlen;
	float *mydirs;

	mydirs = cmdirs;
	
	unsigned long index = 0;
	for(i=0; i<6; i++)
	{
		for(y=0; y<(int)RF_LIGHTRES; y++)
		{
			tc = 2.0f * (y + 0.5f) / RF_LIGHTRES - 1.0f;	// [-1, 1]
			for(x=0; x<(int)RF_LIGHTRES; x++)
			{
				sc = 2.0f * (x + 0.5f) / RF_LIGHTRES - 1.0f;	// [-1, 1]
				switch(i)	// order px, nx, py, ny, pz, nz
				{
				case 0: v[0] =  1.0f;	v[1] = -tc;		v[2] = -sc;		break;
				case 1: v[0] = -1.0f;	v[1] = -tc;		v[2] =  sc;		break;
				case 2: v[0] =  sc;		v[1] =  1.0f;	v[2] =  tc;		break;
				case 3: v[0] =  sc;		v[1] = -1.0f;	v[2] = -tc;		break;
				case 4: v[0] =  sc;		v[1] = -tc;		v[2] =  1.0f;	break;
				case 5: v[0] = sc;		v[1] = tc;		v[2] = -1.0f;	break;
				}
				len = (float) sqrt(v[0]*v[0] + v[1]*v[1] + v[2]*v[2]);
				weightsvector[index] = ((4.0 / RF_BLOCKS) / (len*len*len));	
				if(normalize)
				{
					invlen = 1.0f / len;
					v[0]*=invlen;
					v[1]*=invlen;
					v[2]*=invlen;
				}
				mydirs[0] = v[0];
				mydirs[1] = v[1];
				mydirs[2] = v[2];
				mydirs+=3;
				index ++;
			}
		}
	}
}
// =================================================================================================================


// =================================================================================================================
//  copyCubeMapDirections
//  copy a cube map directions vector to another one
// -----------------------------------------------------------------------------------------------------------------
void copyCubeMapDirections(float *destination, float *source)
{
	unsigned long i;
	for(i=0;i<(RF_LIGHTRES*RF_LIGHTRES*3*6);i++)
		destination[i] = source[i];
}
// =================================================================================================================


// =================================================================================================================
//  multAndGetTexelCubeMapDirections
//  multiply each direction on the vector by the given rotation matrix (results on other vector)
//  resulting vertices are normalized or not depending on the given parameter.
//  Get the corresponding texels values using the other parameters.
// -----------------------------------------------------------------------------------------------------------------
//                                    IN             OUT               IN                  IN              OUT                    IN            IN            IN            IN            IN            IN            IN
void multAndGetTexelCubeMapDirections(float *cmdirs, float *rotcmdirs, float matrix[4][4], bool normalize, double (*texelvec)[3], float *face0, float *face1, float *face2, float *face3, float *face4, float *face5, int facesize)
{
	int i;
	int x, y;
	float vec[3];
	float vabs[3];
	float len, invlen;
	float *mydirs;
	float *myoridirs;
	double (*mytvec)[3];
	int face;
	float facemult;
	unsigned long index;

	int vtex[3];
	int u, v;

	myoridirs = cmdirs;
	mydirs = rotcmdirs;
	mytvec = texelvec;
	
	for(i=0; i<6; i++)
	{
		for(y=0; y<(int)RF_LIGHTRES; y++)
		{
			for(x=0; x<(int)RF_LIGHTRES; x++)
			{
				// -----------------------------------------------------------------------------------
				// MULTIPLY
				vec[0] = matrix[0][0] * myoridirs[0] + matrix[1][0] * myoridirs[1] + matrix[2][0] * myoridirs[2];
				vec[1] = matrix[0][1] * myoridirs[0] + matrix[1][1] * myoridirs[1] + matrix[2][1] * myoridirs[2];
				vec[2] = matrix[0][2] * myoridirs[0] + matrix[1][2] * myoridirs[1] + matrix[2][2] * myoridirs[2];

				if(normalize)
				{
					len = (float) sqrt(vec[0]*vec[0] + vec[1]*vec[1] + vec[2]*vec[2]);
					invlen = 1.0f / len;
					vec[0]*=invlen;
					vec[1]*=invlen;
					vec[2]*=invlen;
				}
				mydirs[0] = vec[0];
				mydirs[1] = vec[1];
				mydirs[2] = vec[2];
				mydirs+=3;
				myoridirs+=3;
				// -----------------------------------------------------------------------------------

				// -----------------------------------------------------------------------------------
				// GET TEXEL
				vabs[0] = (float) fabs(vec[0]);
				vabs[1] = (float) fabs(vec[1]);
				vabs[2] = (float) fabs(vec[2]);

				if(vabs[0] >= vabs[1] && vabs[0] >= vabs[2])	// +-X
				{
					if(vec[0] > 0)	// +X
					{
						face = 0;
					} else {		// -X
						face = 1;
					}
					facemult = 1.0f / vabs[0];
				}
				if(vabs[1] > vabs[0] && vabs[1] >= vabs[2])		// +-Y
				{
					if(vec[1] > 0)	// +Y
					{
						face = 2;
					} else {		// -Y
						face = 3;
					}
					facemult = 1.0f / vabs[1];
				}
				if(vabs[2] > vabs[0] && vabs[2] > vabs[1])		// +-Z
				{
					if(vec[2] > 0)	// +Z
					{
						face = 4;
					} else {		// -Z
						face = 5;
					}
					facemult = 1.0f / vabs[2];
				}

				vabs[0] = (float) ((vec[0]*facemult+1.0)*0.5);
				vabs[1] = (float) ((vec[1]*facemult+1.0)*0.5);
				vabs[2] = (float) ((vec[2]*facemult+1.0)*0.5);

				vtex[0] = (int)(vabs[0] * facesize);
				vtex[1] = (int)(vabs[1] * facesize);
				vtex[2] = (int)(vabs[2] * facesize);

				switch(face)
				{
				case 0:	// +X
					u = vtex[2];
					v = vtex[1];
					index = 3*(v * facesize + u);
					mytvec[0][0] = face0[index];
					mytvec[0][1] = face0[index+1];
					mytvec[0][2] = face0[index+2];
					break;
				case 1:	// -X
					u = vtex[2];
					v = vtex[1];
					index = 3*(v * facesize + u);
					mytvec[0][0] = face1[index];
					mytvec[0][1] = face1[index+1];
					mytvec[0][2] = face1[index+2];
					break;
				case 2:	// +Y
					u = vtex[0];
					v = vtex[2];
					index = 3*(v * facesize + u);
					mytvec[0][0] = face2[index];
					mytvec[0][1] = face2[index+1];
					mytvec[0][2] = face2[index+2];
					break;
				case 3:	// -Y
					u = vtex[0];
					v = vtex[2];
					index = 3*(v * facesize + u);
					mytvec[0][0] = face3[index];
					mytvec[0][1] = face3[index+1];
					mytvec[0][2] = face3[index+2];
					break;
				case 4:	// +Z
					u = vtex[0];
					v = vtex[1];
					index = 3*(v * facesize + u);
					mytvec[0][0] = face4[index];
					mytvec[0][1] = face4[index+1];
					mytvec[0][2] = face4[index+2];
					break;
				case 5:	// -Z
					u = vtex[0];
					v = vtex[1];
					index = 3*(v * facesize + u);
					mytvec[0][0] = face5[index];
					mytvec[0][1] = face5[index+1];
					mytvec[0][2] = face5[index+2];
					break;
				}
				mytvec[0][0]*=weights[i*RF_LIGHTRES*RF_LIGHTRES + y*RF_LIGHTRES + x];
				mytvec[0][1]*=weights[i*RF_LIGHTRES*RF_LIGHTRES + y*RF_LIGHTRES + x];
				mytvec[0][2]*=weights[i*RF_LIGHTRES*RF_LIGHTRES + y*RF_LIGHTRES + x];
				s[i*RF_LIGHTRES*RF_LIGHTRES + y*RF_LIGHTRES + x][0] = mytvec[0][0] * exposure;
				s[i*RF_LIGHTRES*RF_LIGHTRES + y*RF_LIGHTRES + x][1] = mytvec[0][1] * exposure;
				s[i*RF_LIGHTRES*RF_LIGHTRES + y*RF_LIGHTRES + x][2] = mytvec[0][2] * exposure;
				mytvec+=3;
				// -----------------------------------------------------------------------------------
			}
		}
	}
}
// =================================================================================================================

// ***************************************************************************************************
// ***************************************************************************************************
