// ***************************************************************************************************
// ***************************************************************************************************
//  RELIGHTING FRAMEWORK v0.10
//  Sebastian Enrique - senrique@cs.columbia.edu
//  Columbia University
//  January 2004
// ***************************************************************************************************
// ***************************************************************************************************

// Features wish list:
// - Load/Save status of light/cubemap
// - Hability to change cubemap in real-time

// =================================================================================================================
//  INCLUDE FILES AND LIBRARIES
// -----------------------------------------------------------------------------------------------------------------
#include <windows.h>											// Header File For Windows
#include <gl\gl.h>												// Header File For The OpenGL32 Library
#include <gl\glu.h>												// Header File For The GLu32 Library
#include <gl\glut.h>											// Header File For The GLu32 Library
#include <gl\glaux.h>											// Header File For The GLaux Library
#include <math.h>	
#include <stdio.h>
#include <string.h>
#include "TGAImage.h"											/* TGA Image loader				 */
#include "trackball.c"											/* trackball quaternion routines */
#include "rgbe.c"												/* HDR Image loader				 */
#include "rfcparser.c"											/* RFC parser					 */
#include "cmcalc.h"												/* Some Cube Map Computations	 */
#include "time.h"

#pragma comment( lib, "opengl32.lib" )							// Search For OpenGL32.lib While Linking
#pragma comment( lib, "glu32.lib" )								// Search For GLu32.lib While Linking
#pragma comment( lib, "glaux.lib" )								// Search For GLaux.lib While Linking
// =================================================================================================================


// =================================================================================================================
//  CONSTANTS DEFINITIONS
// -----------------------------------------------------------------------------------------------------------------
// SAVE VIDEO FEATURE
#define RF_FREQUENCY		1000.0 / 30.0	// Clock Frequency (TICKS PER SECONDS OVER FRAMES PER SECONDS)
#define RF_VIDEOFILE		"video.rfv"		// Rotation Matrices output file
#define RF_VIDEODIR			"video\\"		// Images output path

// SCREEN RESOLUTION
#define	SCREEN_WIDTH	924	// 768	//1024
#define SCREEN_HEIGHT	693	// 576	//768

// CUBEMAP TEXTURE EXTENSIONS
#ifndef GL_ARB_texture_cube_map 
#define GL_ARB_texture_cube_map             1 
#define GL_NORMAL_MAP_ARB                   0x8511 
#define GL_REFLECTION_MAP_ARB               0x8512 
#define GL_TEXTURE_CUBE_MAP_ARB             0x8513 
#define GL_TEXTURE_BINDING_CUBE_MAP_ARB     0x8514 
#define GL_TEXTURE_CUBE_MAP_POSITIVE_X_ARB  0x8515 
#define GL_TEXTURE_CUBE_MAP_NEGATIVE_X_ARB  0x8516 
#define GL_TEXTURE_CUBE_MAP_POSITIVE_Y_ARB  0x8517 
#define GL_TEXTURE_CUBE_MAP_NEGATIVE_Y_ARB  0x8518 
#define GL_TEXTURE_CUBE_MAP_POSITIVE_Z_ARB  0x8519 
#define GL_TEXTURE_CUBE_MAP_NEGATIVE_Z_ARB  0x851A 
#define GL_PROXY_TEXTURE_CUBE_MAP_ARB       0x851B 
#define GL_MAX_CUBE_MAP_TEXTURE_SIZE_ARB    0x851C 
#endif

// POINT LIGHTS
#define RF_MAXLIGHTS	10		// Max number of point lights
#define RF_DELTA_COLOR	0.05	// Step of incrementing/decrementing point light color component
#define RF_SCALE_FACTOR	2.0		// Display setting - (each point light is displayed in the grid with a side of 2 times its value)

// MODES ENUMERATION
#define RF_TEX_NOTEXTURE	0	// Point Lights
#define RF_TEX_CUBEMAP		1	// Cubic Environment Map
#define RF_TEX_POINTCUBE	2	// Point Lights + Cubic Environment Map

// CUBIC ENVIRONMENT MAP DISPLAY CONSTANTS
#define	RF_CRUX_XC	75.0f		// Position for the center of the front face in x axis [0..100]
#define RF_CRUX_YC	15.0f		// Position for the top of the front face in y aixs [0..100]
#define RF_CRUX_DELTA	9.0f	// Lenght of each side

// CUBEMAP
#define RF_MAX_CUBEMAP_SIZE	256*256		// Max allowed cubemap resolution

// POINT LIGHTS DISPLAY CONSTANTS
#define RF_LIGHTSPANEL_X0		85.0f	// X-axis position for the point lights panel (light 0)
#define RF_LIGHTSPANEL_Y0		85.0f	// Y-axis position for the point lights panel (light 0)
#define RF_LIGHTSPANEL_SCALE	1.5f	// Scale factor for the point lights panel
#define RF_LIGHTSPANEL_YSTEP	2.0f	// Step between each light panel information
// =================================================================================================================


// =================================================================================================================
//  STRUCTURES
// -----------------------------------------------------------------------------------------------------------------
struct Light {				// POINT LIGHT
	bool	on;				// Light is on or off
	unsigned int xIndex;	// X position inside grid
	unsigned int yIndex;	// Y position inside grid
	double color[3];		// Color
	unsigned char scale;	// Scale: 1, 1 grid block; 2, 3x3 grid block; 3, 5x5 grid block and so on.
};
// =================================================================================================================


// =================================================================================================================
//  GLOBAL VARIABLES
// -----------------------------------------------------------------------------------------------------------------
// INPUT VALUES
unsigned int RF_IMAGERESW;	// Image width in pixels
unsigned int RF_IMAGERESH;	// Image height in pixels
unsigned int RF_LIGHTRES;	// How many squared lights for each face are being used
unsigned int RF_BLOCKRESW;	// Blocks in the image width
unsigned int RF_BLOCKRESH;	// Blocks in the image height
unsigned int RF_PIXELRES;	// How many squared pixels for each block
unsigned long RF_IMAGES;	// RF_LIGHTRES * RF_LIGHTRES * 6
unsigned long RF_BLOCKS;	// RF_BLOCKRESW * RF_BLOCKRESH
unsigned long RF_PIXELS;	// RF_PIXELRES * RF_PIXELRES
unsigned int RF_RANK_B;		// Rank used in the first PCA
unsigned int RF_RANK_Q;		// Rank used in the second PCA

// INPUT FILES
char* RF_HDRFILE;			// High Dynamic Range path + file name (without .hdr extension; a tga file with the same name should exist in the same path -this is used for the tone map-)
char* RF_ERFILE;			// E red component matrix path + name
char* RF_EGFILE;			// E green component matrix path + name
char* RF_EBFILE;			// E blue component matrix path + name
char* RF_VRFILE;			// U red component matrix path + name
char* RF_VGFILE;			// U green component matrix path + name
char* RF_VBFILE;			// U blue component matrix path + name
char* RF_URFILE;			// V red component matrix path + name
char* RF_UGFILE;			// V green component matrix path + name
char* RF_UBFILE;			// V blue component matrix path + name

// CURRENT SCREEN RESOLUTION
int sw, sh;					// sw: width; sh: height

// VECTORS AND MATRICES
double ***Er;				// E red component matrix
double ***Eg;				// E green component matrix
double ***Eb;				// E blue component matrix
double **Vr;				// V red component matrix
double **Vg;				// V green component matrix
double **Vb;				// V blue component matrix
double **Ur;				// U red component matrix
double **Ug;				// U green component matrix
double **Ub;				// U blue component matrix
double (*s)[3];				// Light Field Vector
double *weights;			// Light Weights (solid angle) Vector

double (*Vs)[3];			// V times s aux vector
double (*UVs)[3];			// U times V times s aux vector
double (*EjUVs)[3];			// Ej times U times V times s vector

float  *I;					// Image matrix

float  *myCMDirs;			// Original CubeMap Directions vector
float  *myRotCMDirs;		// Rotated CubeMap Directions vector

// OTHER AUXILIARY VARIABLES
char straux[1024];			// For strings generation
bool bChange;				// true if image should be recomputated
bool bFullScreen;			// Indicates full screen / windowed mode
bool bToneMap;				// See ToneMap / Interpolated HDR
unsigned char texFlag;		// Current Display Mode

// POINT LIGHTS
char selectedLight;							// Current selected point light
unsigned char selectedLightQty;				// How many point lights are turned on
Light myLights[RF_MAXLIGHTS];				// Point lights vector
unsigned char lightPanelKey[RF_MAXLIGHTS];	// Keys to select/turn on/off each point light
int lightPanelSel[RF_MAXLIGHTS][4];			// Bounding Box coordinates to select/turn on/off light
int lightPanelColor[RF_MAXLIGHTS][6][4];	// Bounding Box coordinates for arrows to increment/decrement colors [R+ R- G+ G- B+ B-]
float xTexCoord, yTexCoord;					// Current Light Coordinate

// TEXTURES
unsigned char* C;				// Texture ???
unsigned int cWidth, cHeight;	// Texture ???

// CUBE
float curquat[4];							// Current Quaternion
float lastquat[4];							// Last Quaternion
int beginx, beginy, cubex, cubey, W, H;		// Several variables useful for quaternion rotations and mouse interface for
int changingcolor, moving;					// rotation and color change

// HDR FILE
FILE* f;						// HDR/TGA File Handler
float* image;					// HDR Image Data
int image_width;				// HDR Image width
int image_height;				// HDR Image height
rgbe_header_info* myHeader;		// HDR/TGA Header

float* image_cube0;				// HDR Right Face interpolated
float* image_cube1;				// HDR Left Face interpolated
float* image_cube2;				// HDR Top Face interpolated
float* image_cube3;				// HDR Bottom Face interpolated
float* image_cube4;				// HDR Front Face interpolated
float* image_cube5;				// HDR Back Face interpolated

TGAImage myTGA;					// TGA Image
unsigned char* image_cube0_tga;	// TGA Right Face (tone map)
unsigned char* image_cube1_tga;	// TGA Left Face (tone map)
unsigned char* image_cube2_tga;	// TGA Top Face (tone map)
unsigned char* image_cube3_tga;	// TGA Bottom Face (tone map)
unsigned char* image_cube4_tga;	// TGA Front Face (tone map)
unsigned char* image_cube5_tga;	// TGA Back Face (tone map)

int oldWidth, oldHeight;
double exposure = 1.0;

// TIMER and OTHER VARIABLES
clock_t lastTime;
clock_t currentTime;
double elapsedTime;
bool saveMatrices = false;
bool saveScreenshot = false;
FILE* myVideoFile = NULL;				// File handler to video file
float matrix[4][4];						// Current Rotation Matrix
bool loadMatrices = false;
FILE* myVideoFileLoad = NULL;
long filenumber = 0;
unsigned char data[2048*2048*3];
unsigned char bgraData[2048*2048*3];
// =================================================================================================================


// =========================================================================================
//  indexI
//  Get the index of the element corresponding to x and y positions with color component z
// -----------------------------------------------------------------------------------------
unsigned long indexI(unsigned long x, unsigned long y, unsigned long z)
{
	return x*RF_IMAGERESW*3 + y*3 + z;
}
// =================================================================================================================


// =========================================================================================
//  Generate Video File
//  Saves the current rotation matrix to the output video file
// -----------------------------------------------------------------------------------------
void genVideo()
{
	if(myVideoFile==NULL)
	{
		myVideoFile = fopen(RF_VIDEOFILE, "a");
	}

	if(myVideoFile!=NULL)
	{
		fprintf(myVideoFile, "%f %f %f %f %f %f %f %f %f\n",matrix[0][0],matrix[0][1],matrix[0][2],matrix[1][0],matrix[1][1],matrix[1][2],matrix[2][0],matrix[2][1],matrix[2][2]);
	}
}
// =========================================================================================


// =========================================================================================
//  Precompute Light Panel Coordinates
// -----------------------------------------------------------------------------------------
void lightPanelPrecomputation(void)
{
	float scale, step;
	int leftx, lefty, rightx, righty;
	int rlx, rrx, glx, grx, blx, brx;
	int caly, cary, csly, csry;
	unsigned int i;

	lightPanelKey[0] = '0';
	lightPanelKey[1] = '1';
	lightPanelKey[2] = '2';
	lightPanelKey[3] = '3';
	lightPanelKey[4] = '4';
	lightPanelKey[5] = '5';
	lightPanelKey[6] = '6';
	lightPanelKey[7] = '7';
	lightPanelKey[8] = '8';
	lightPanelKey[9] = '9';

	scale = RF_LIGHTSPANEL_SCALE * (100.0f / (float)sh);

	step = 0.0f;
	leftx = (int)(RF_LIGHTSPANEL_X0 * (float)sw / 100.0f);
	rightx = (int)((RF_LIGHTSPANEL_X0+14.0f*scale) * (float)sw / 100.0f);
	rlx = (int)((RF_LIGHTSPANEL_X0+26.0f*scale) * (float)sw / 100.0f);;
	rrx = (int)((RF_LIGHTSPANEL_X0+32.0f*scale) * (float)sw / 100.0f);;
	glx = (int)((RF_LIGHTSPANEL_X0+34.0f*scale) * (float)sw / 100.0f);;
	grx = (int)((RF_LIGHTSPANEL_X0+40.0f*scale) * (float)sw / 100.0f);;
	blx = (int)((RF_LIGHTSPANEL_X0+42.0f*scale) * (float)sw / 100.0f);;
	brx = (int)((RF_LIGHTSPANEL_X0+48.0f*scale) * (float)sw / 100.0f);;
	for(i=0;i<10;i++)
	{
		lefty = sh-(int)((RF_LIGHTSPANEL_Y0-step+4.0f*scale) * (float)sh / 100.0f);
		righty = sh-(int)((RF_LIGHTSPANEL_Y0-step) * (float)sh / 100.0f);
		caly = sh-(int)((RF_LIGHTSPANEL_Y0-step+7.0f*scale) * (float)sh / 100.0f);
		cary = sh-(int)((RF_LIGHTSPANEL_Y0-step+3.0f*scale) * (float)sh / 100.0f);
		csly = sh-(int)((RF_LIGHTSPANEL_Y0-step+1.0f*scale) * (float)sh / 100.0f);
		csry = sh-(int)((RF_LIGHTSPANEL_Y0-step-3.0f*scale) * (float)sh / 100.0f);
		lightPanelSel[i][0] = leftx;
		lightPanelSel[i][1] = rightx;
		lightPanelSel[i][2] = lefty;
		lightPanelSel[i][3] = righty;
		lightPanelColor[i][0][0] = rlx;
		lightPanelColor[i][0][1] = rrx;
		lightPanelColor[i][0][2] = caly;
		lightPanelColor[i][0][3] = cary;
		lightPanelColor[i][1][0] = rlx;
		lightPanelColor[i][1][1] = rrx;
		lightPanelColor[i][1][2] = csly;
		lightPanelColor[i][1][3] = csry;
		lightPanelColor[i][2][0] = glx;
		lightPanelColor[i][2][1] = grx;
		lightPanelColor[i][2][2] = caly;
		lightPanelColor[i][2][3] = cary;
		lightPanelColor[i][3][0] = glx;
		lightPanelColor[i][3][1] = grx;
		lightPanelColor[i][3][2] = csly;
		lightPanelColor[i][3][3] = csry;
		lightPanelColor[i][4][0] = blx;
		lightPanelColor[i][4][1] = brx;
		lightPanelColor[i][4][2] = caly;
		lightPanelColor[i][4][3] = cary;
		lightPanelColor[i][5][0] = blx;
		lightPanelColor[i][5][1] = brx;
		lightPanelColor[i][5][2] = csly;
		lightPanelColor[i][5][3] = csry;
		step += RF_LIGHTSPANEL_YSTEP;
	}
}
// =========================================================================================


// =========================================================================================
//  Custom Initialization for Relighting
// -----------------------------------------------------------------------------------------
void MyInit(void)
{
	unsigned int ii,jj;

	bFullScreen = false;	// Initially in Windowed Mode

	RF_IMAGERESW = myRFCData.imageresw;
	RF_IMAGERESH = myRFCData.imageresh;
	RF_LIGHTRES  = myRFCData.lightres;
	RF_BLOCKRESW = myRFCData.blockresw;
	RF_BLOCKRESH = myRFCData.blockresh;
	RF_PIXELRES  = myRFCData.pixelres;
	RF_IMAGES    = RF_LIGHTRES*RF_LIGHTRES*6;	// n
	RF_BLOCKS    = RF_BLOCKRESW*RF_BLOCKRESH;	// m
	RF_PIXELS    = RF_PIXELRES*RF_PIXELRES;		// p
	RF_RANK_B    = myRFCData.rankb;
	RF_RANK_Q    = myRFCData.rankq;

	weights = new double[RF_IMAGES];

	myCMDirs = (float *)malloc(sizeof(float) * RF_LIGHTRES * RF_LIGHTRES * 3 * 6);
	myRotCMDirs = (float *)malloc(sizeof(float) * RF_LIGHTRES * RF_LIGHTRES * 3 * 6);
	
	printf("Generating CubeMap Directions\n");

	generateCubeMapDirections(myCMDirs, false, weights);
	copyCubeMapDirections(myRotCMDirs, myCMDirs);

	Er = new double**[RF_BLOCKS];
	for(ii=0; ii<RF_BLOCKS; ii++)
	{
		Er[ii] = new double*[RF_PIXELS];
		for(jj=0; jj<RF_PIXELS; jj++)
			Er[ii][jj] = new double[RF_RANK_B];
	}

	Eg = new double**[RF_BLOCKS];
	for(ii=0; ii<RF_BLOCKS; ii++)
	{
		Eg[ii] = new double*[RF_PIXELS];
		for(jj=0; jj<RF_PIXELS; jj++)
			Eg[ii][jj] = new double[RF_RANK_B];
	}

	Eb = new double**[RF_BLOCKS];
	for(ii=0; ii<RF_BLOCKS; ii++)
	{
		Eb[ii] = new double*[RF_PIXELS];
		for(jj=0; jj<RF_PIXELS; jj++)
			Eb[ii][jj] = new double[RF_RANK_B];
	}

	Vr = new double*[RF_RANK_Q];
	for(ii=0; ii<RF_RANK_Q; ii++)
		Vr[ii] = new double[RF_IMAGES];

	Vg = new double*[RF_RANK_Q];
	for(ii=0; ii<RF_RANK_Q; ii++)
		Vg[ii] = new double[RF_IMAGES];

	Vb = new double*[RF_RANK_Q];
	for(ii=0; ii<RF_RANK_Q; ii++)
		Vb[ii] = new double[RF_IMAGES];

	Ur = new double*[RF_BLOCKS*RF_RANK_B];
	for(ii=0; ii<(RF_BLOCKS*RF_RANK_B); ii++)
		Ur[ii] = new double[RF_RANK_Q];

	Ug = new double*[RF_BLOCKS*RF_RANK_B];
	for(ii=0; ii<(RF_BLOCKS*RF_RANK_B); ii++)
		Ug[ii] = new double[RF_RANK_Q];

	Ub = new double*[RF_BLOCKS*RF_RANK_B];
	for(ii=0; ii<(RF_BLOCKS*RF_RANK_B); ii++)
		Ub[ii] = new double[RF_RANK_Q];

	s = new double[RF_IMAGES][3];
	
	Vs = new double[RF_RANK_Q][3];
	UVs = new double[RF_BLOCKS*RF_RANK_B][3];

	I = new float[RF_IMAGERESW*RF_IMAGERESH*3];

	EjUVs = new double[RF_PIXELS][3];

	C = new unsigned char[RF_MAX_CUBEMAP_SIZE*4];

	printf("Loading High Dynamic Range File\n");

	// HDR FILE
	myHeader = new rgbe_header_info;
	sprintf(straux, "%s.hdr", myRFCData.hdrfile);
	f = fopen(straux,"rb");
	if(f==NULL)
	{
		printf("ERROR: HDR File %s cannot be found. Exiting.\n",myRFCData.hdrfile);
		exit(1);
	}
	RGBE_ReadHeader(f,&image_width,&image_height,myHeader);
	image = (float *)malloc(sizeof(float)*3*image_width*image_height);
	RGBE_ReadPixels_RLE(f,image,image_width,image_height);
	fclose(f);

	image_cube0 = (float *)malloc(sizeof(float)*3*(image_width/3)*(image_height/4));
	image_cube1 = (float *)malloc(sizeof(float)*3*(image_width/3)*(image_height/4));
	image_cube2 = (float *)malloc(sizeof(float)*3*(image_width/3)*(image_height/4));
	image_cube3 = (float *)malloc(sizeof(float)*3*(image_width/3)*(image_height/4));
	image_cube4 = (float *)malloc(sizeof(float)*3*(image_width/3)*(image_height/4));
	image_cube5 = (float *)malloc(sizeof(float)*3*(image_width/3)*(image_height/4));

	int patchsize = 2*((image_height/4)/RF_LIGHTRES);
	RGBE_GetInterpolatedFace(0, image, image_cube0, image_width, image_height, patchsize);
	RGBE_GetInterpolatedFace(1, image, image_cube1, image_width, image_height, patchsize);
	RGBE_GetInterpolatedFace(2, image, image_cube2, image_width, image_height, patchsize);
	RGBE_GetInterpolatedFace(3, image, image_cube3, image_width, image_height, patchsize);
	RGBE_GetInterpolatedFace(4, image, image_cube4, image_width, image_height, patchsize);
	RGBE_GetInterpolatedFace(5, image, image_cube5, image_width, image_height, patchsize);

	delete myHeader;

	// TGA FILE
	myHeader = new rgbe_header_info;
	sprintf(straux, "%s.tga", myRFCData.hdrfile);
	f = fopen(straux,"rb");
	if(f==NULL)
	{
		printf("ERROR: TGA File %s cannot be found. Exiting.\n",myRFCData.hdrfile);
		exit(1);
	}
	fclose(f);
	myTGA.open(straux);

	image_cube0_tga = (unsigned char *)malloc(sizeof(unsigned char)*3*(image_width/3)*(image_height/4));
	image_cube1_tga = (unsigned char *)malloc(sizeof(unsigned char)*3*(image_width/3)*(image_height/4));
	image_cube2_tga = (unsigned char *)malloc(sizeof(unsigned char)*3*(image_width/3)*(image_height/4));
	image_cube3_tga = (unsigned char *)malloc(sizeof(unsigned char)*3*(image_width/3)*(image_height/4));
	image_cube4_tga = (unsigned char *)malloc(sizeof(unsigned char)*3*(image_width/3)*(image_height/4));
	image_cube5_tga = (unsigned char *)malloc(sizeof(unsigned char)*3*(image_width/3)*(image_height/4));

	myTGA.getFace(0, image_cube0_tga);
	myTGA.getFace(1, image_cube1_tga);
	myTGA.getFace(2, image_cube2_tga);
	myTGA.getFace(3, image_cube3_tga);
	myTGA.getFace(4, image_cube4_tga);
	myTGA.getFace(5, image_cube5_tga);


	sw = SCREEN_WIDTH;
	sh = SCREEN_HEIGHT;

	unsigned int i, j, k;
	FILE *fpR, *fpG, *fpB;

	// s initialization
	printf("Initializing Lighting Vector\n");
	for(i=0;i<RF_IMAGES;i++)
		s[i][0] = s[i][1] = s[i][2] = 0;

	// load matrices E
	printf("Loading E matrices\n");
	fpR = fopen(myRFCData.erfile,"r");
	if(fpR==NULL)
	{
		printf("ERROR: Er Matrix File %s cannot be found. Exiting.\n",myRFCData.erfile);
		exit(1);
	}
	fpG = fopen(myRFCData.egfile,"r");
	if(fpG==NULL)
	{
		fclose(fpR);
		printf("ERROR: Eg Matrix File %s cannot be found. Exiting.\n",myRFCData.egfile);
		exit(1);
	}
	fpB = fopen(myRFCData.ebfile,"r");
	if(fpB==NULL)
	{
		fclose(fpR);
		fclose(fpG);
		printf("ERROR: Eb Matrix File %s cannot be found. Exiting.\n",myRFCData.ebfile);
		exit(1);
	}
	for(k=0; k<RF_BLOCKS; k++)
		for(i=0; i<RF_PIXELS; i++)
			for(j=0; j<RF_RANK_B; j++)
			{
				 fscanf(fpR, " %lf ", &Er[k][i][j]);
				 fscanf(fpG, " %lf ", &Eg[k][i][j]);
				 fscanf(fpB, " %lf ", &Eb[k][i][j]);
			}
	fclose(fpR);
	fclose(fpG);
	fclose(fpB);

	// load matrices V
	printf("Loading V matrices\n");
	fpR = fopen(myRFCData.vrfile,"r");
	if(fpR==NULL)
	{
		printf("ERROR: Vr Matrix File %s cannot be found. Exiting.\n",myRFCData.vrfile);
		exit(1);
	}
	fpG = fopen(myRFCData.vgfile,"r");
	if(fpG==NULL)
	{
		fclose(fpR);
		printf("ERROR: Vg Matrix File %s cannot be found. Exiting.\n",myRFCData.vgfile);
		exit(1);
	}
	fpB = fopen(myRFCData.vbfile,"r");
	if(fpB==NULL)
	{
		fclose(fpR);
		fclose(fpG);
		printf("ERROR: Vb Matrix File %s cannot be found. Exiting.\n",myRFCData.vbfile);
		exit(1);
	}
	for(i=0; i<RF_RANK_Q; i++)
		for(j=0; j<RF_IMAGES; j++)
		{
			 fscanf(fpR, " %lf ", &Vr[i][j]);
			 fscanf(fpG, " %lf ", &Vg[i][j]);
			 fscanf(fpB, " %lf ", &Vb[i][j]);
		}
	fclose(fpR);
	fclose(fpG);
	fclose(fpB);

	// load matrices U
	printf("Loading U matrices\n");
	fpR = fopen(myRFCData.urfile,"r");
	if(fpR==NULL)
	{
		printf("ERROR: Ur Matrix File %s cannot be found. Exiting.\n",myRFCData.urfile);
		exit(1);
	}
	fpG = fopen(myRFCData.ugfile,"r");
	if(fpG==NULL)
	{
		fclose(fpR);
		printf("ERROR: Ug Matrix File %s cannot be found. Exiting.\n",myRFCData.ugfile);
		exit(1);
	}
	fpB = fopen(myRFCData.ubfile,"r");
	if(fpB==NULL)
	{
		fclose(fpR);
		fclose(fpG);
		printf("ERROR: Ub Matrix File %s cannot be found. Exiting.\n",myRFCData.ubfile);
		exit(1);
	}
	for(i=0; i<RF_BLOCKS*RF_RANK_B; i++)
		for(j=0; j<RF_RANK_Q; j++)
		{
			 fscanf(fpR, " %lf ", &Ur[i][j]);
			 fscanf(fpG, " %lf ", &Ug[i][j]);
			 fscanf(fpB, " %lf ", &Ub[i][j]);
		}
	fclose(fpR);
	fclose(fpG);
	fclose(fpB);

	printf("Performing final initialization\n");

	// point light vector inicialization
	for(i=0;i<RF_MAXLIGHTS;i++)
	{
		myLights[i].color[0] = 1.0;
		myLights[i].color[1] = 1.0;
		myLights[i].color[2] = 1.0;
		myLights[i].xIndex = 0;
		myLights[i].yIndex = 0;
		myLights[i].scale = 1;
		myLights[i].on = false;
	}

	// setting first light on
	myLights[0].on = true;
	selectedLight = 0;
	selectedLightQty = 1;
	s[4*RF_LIGHTRES*RF_LIGHTRES+RF_LIGHTRES*(RF_LIGHTRES - myLights[selectedLight].yIndex - 1) + myLights[selectedLight].xIndex][0] = myLights[selectedLight].color[0];
	s[4*RF_LIGHTRES*RF_LIGHTRES+RF_LIGHTRES*(RF_LIGHTRES - myLights[selectedLight].yIndex - 1) + myLights[selectedLight].xIndex][1] = myLights[selectedLight].color[1];
	s[4*RF_LIGHTRES*RF_LIGHTRES+RF_LIGHTRES*(RF_LIGHTRES - myLights[selectedLight].yIndex - 1) + myLights[selectedLight].xIndex][2] = myLights[selectedLight].color[2];

	// recompute image
	bChange = true;

	// GL initialization
	glShadeModel(GL_FLAT);
	glPixelStorei(GL_UNPACK_ALIGNMENT, 1);

	// Texture Mapping
	texFlag = RF_TEX_NOTEXTURE;
	xTexCoord = yTexCoord = 0.0f;

	glTexParameteri(GL_TEXTURE_CUBE_MAP_ARB, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_CUBE_MAP_ARB, GL_TEXTURE_MIN_FILTER, GL_LINEAR);

	// TONE MAP
	glTexImage2D(GL_TEXTURE_CUBE_MAP_POSITIVE_X_ARB, 0, GL_RGB, image_height/4, image_width/3, 0, GL_RGB, GL_UNSIGNED_BYTE, image_cube0_tga);	// Height, Width
	glTexImage2D(GL_TEXTURE_CUBE_MAP_NEGATIVE_X_ARB, 0, GL_RGB, image_height/4, image_width/3, 0, GL_RGB, GL_UNSIGNED_BYTE, image_cube1_tga);	// &myTexMap2.texCubeData[0] + RF_LIGHTRES*RF_LIGHTRES*3);	// Height, Width
	glTexImage2D(GL_TEXTURE_CUBE_MAP_POSITIVE_Y_ARB, 0, GL_RGB, image_height/4, image_width/3, 0, GL_RGB, GL_UNSIGNED_BYTE, image_cube2_tga);	// Height, Width
	glTexImage2D(GL_TEXTURE_CUBE_MAP_NEGATIVE_Y_ARB, 0, GL_RGB, image_height/4, image_width/3, 0, GL_RGB, GL_UNSIGNED_BYTE, image_cube3_tga);	// Height, Width
	glTexImage2D(GL_TEXTURE_CUBE_MAP_POSITIVE_Z_ARB, 0, GL_RGB, image_height/4, image_width/3, 0, GL_RGB, GL_UNSIGNED_BYTE, image_cube4_tga);	// Height, Width
	glTexImage2D(GL_TEXTURE_CUBE_MAP_NEGATIVE_Z_ARB, 0, GL_RGB, image_height/4, image_width/3, 0, GL_RGB, GL_UNSIGNED_BYTE, image_cube5_tga);	// Height, Width
	bToneMap = true;

	glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_REFLECTION_MAP_ARB);
	glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_REFLECTION_MAP_ARB);
	glTexGeni(GL_R, GL_TEXTURE_GEN_MODE, GL_REFLECTION_MAP_ARB);
	glTexParameteri(GL_TEXTURE_CUBE_MAP_ARB, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_CUBE_MAP_ARB, GL_TEXTURE_WRAP_T, GL_REPEAT);
	texFlag = RF_TEX_CUBEMAP;

	cubex = cubey = 0;
	trackball(curquat, 0.0, 0.0, 0.0, 0.0);

	lightPanelPrecomputation();
	changingcolor = -1;

	printf("Ready!\n");
}
// =========================================================================================


// =========================================================================================
//  Substract a light from the scene
// -----------------------------------------------------------------------------------------
void subLight(unsigned char number)
{
	int i, j, delta, size;
	
	delta = 1 - myLights[number].scale;
	size = 2 * myLights[number].scale - 1;

	for(i=myLights[number].xIndex + delta; i<(int)(myLights[number].xIndex + delta + size); i++)
		for(j=myLights[number].yIndex + delta; j<(int)(myLights[number].yIndex + delta + size); j++)
			if(i>=0 && i<(int)RF_LIGHTRES && j>=0 && j<(int)RF_LIGHTRES)
			{
				// This is only in the front face
				s[RF_LIGHTRES*RF_LIGHTRES*4 + RF_LIGHTRES*(RF_LIGHTRES - j - 1) + i][0]-=myLights[number].color[0];
				s[RF_LIGHTRES*RF_LIGHTRES*4 + RF_LIGHTRES*(RF_LIGHTRES - j - 1) + i][1]-=myLights[number].color[1];
				s[RF_LIGHTRES*RF_LIGHTRES*4 + RF_LIGHTRES*(RF_LIGHTRES - j - 1) + i][2]-=myLights[number].color[2];

				/*
				// To USE LEFT FACE
				s[RF_LIGHTRES*RF_LIGHTRES*1 + RF_LIGHTRES*(RF_LIGHTRES - j - 1) + i][0]-=myLights[number].color[0];
				s[RF_LIGHTRES*RF_LIGHTRES*1 + RF_LIGHTRES*(RF_LIGHTRES - j - 1) + i][1]-=myLights[number].color[1];
				s[RF_LIGHTRES*RF_LIGHTRES*1 + RF_LIGHTRES*(RF_LIGHTRES - j - 1) + i][2]-=myLights[number].color[2];
				*/
			}
}
// =========================================================================================


// =========================================================================================
//  Add a light to the scene
// -----------------------------------------------------------------------------------------
void addLight(unsigned char number)
{
	int i, j, delta, size;
	
	delta = 1 - myLights[number].scale;
	size = 2 * myLights[number].scale - 1;

	for(i=myLights[number].xIndex + delta; i<(int)(myLights[number].xIndex + delta + size); i++)
		for(j=myLights[number].yIndex + delta; j<(int)(myLights[number].yIndex + delta + size); j++)
			if(i>=0 && i<(int)RF_LIGHTRES && j>=0 && j<(int)RF_LIGHTRES)
			{
				// This is only in the front face
				s[RF_LIGHTRES*RF_LIGHTRES*4 + RF_LIGHTRES*(RF_LIGHTRES - j - 1) + i][0]+=myLights[number].color[0];
				s[RF_LIGHTRES*RF_LIGHTRES*4 + RF_LIGHTRES*(RF_LIGHTRES - j - 1) + i][1]+=myLights[number].color[1];
				s[RF_LIGHTRES*RF_LIGHTRES*4 + RF_LIGHTRES*(RF_LIGHTRES - j - 1) + i][2]+=myLights[number].color[2];
				
				/*
				// TO USE LEFT FACE
				s[RF_LIGHTRES*RF_LIGHTRES*1 + RF_LIGHTRES*(RF_LIGHTRES - j - 1) + i][0]+=myLights[number].color[0];
				s[RF_LIGHTRES*RF_LIGHTRES*1 + RF_LIGHTRES*(RF_LIGHTRES - j - 1) + i][1]+=myLights[number].color[1];
				s[RF_LIGHTRES*RF_LIGHTRES*1 + RF_LIGHTRES*(RF_LIGHTRES - j - 1) + i][2]+=myLights[number].color[2];
				*/
			}
}
// =========================================================================================


// =========================================================================================
//  Select / Turn On/Off Light
// -----------------------------------------------------------------------------------------
void checkLight(unsigned char number)
{
	int j;

	if(myLights[number].on)
	{
		if (selectedLight==number)
		{
			if(selectedLightQty>0)
			{
				subLight(number);
				myLights[number].on = false;
				selectedLightQty--;
				j = number;
				while(selectedLight==number && selectedLightQty>0)
				{
					if(j>=RF_MAXLIGHTS)	j-=RF_MAXLIGHTS;
					if(myLights[j].on) selectedLight = j;
					j ++;
				}
				if(selectedLightQty==0) selectedLight = -1;
				bChange = true;
			}
		} else {
			selectedLight = number;
		}
	} else {
		addLight(number);
		myLights[number].on = true;
		selectedLight = number;
		selectedLightQty++;
		bChange = true;
	}
}
// =========================================================================================


// =========================================================================================
//  Keys Capturing
// -----------------------------------------------------------------------------------------
void keyboard(unsigned char c, int x, int y)
{
	unsigned int aux;
	int i;

	if (c == 27)
	{
		exit(0);	// ESC : Exit Program
	}

	if (c == 32)	// SPACE BAR: Switch between TONE MAP / INTERPOLATED HDR MAP
	{
		glTexParameteri(GL_TEXTURE_CUBE_MAP_ARB, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_CUBE_MAP_ARB, GL_TEXTURE_MIN_FILTER, GL_LINEAR);

		if(bToneMap)
		{
			//INTERPOLATED CUBEMAP
			glTexImage2D(GL_TEXTURE_CUBE_MAP_POSITIVE_X_ARB, 0, GL_RGB, image_height/4, image_width/3, 0, GL_RGB, GL_FLOAT, image_cube0);
			glTexImage2D(GL_TEXTURE_CUBE_MAP_NEGATIVE_X_ARB, 0, GL_RGB, image_height/4, image_width/3, 0, GL_RGB, GL_FLOAT, image_cube1);
			glTexImage2D(GL_TEXTURE_CUBE_MAP_POSITIVE_Y_ARB, 0, GL_RGB, image_height/4, image_width/3, 0, GL_RGB, GL_FLOAT, image_cube2);
			glTexImage2D(GL_TEXTURE_CUBE_MAP_NEGATIVE_Y_ARB, 0, GL_RGB, image_height/4, image_width/3, 0, GL_RGB, GL_FLOAT, image_cube3);
			glTexImage2D(GL_TEXTURE_CUBE_MAP_POSITIVE_Z_ARB, 0, GL_RGB, image_height/4, image_width/3, 0, GL_RGB, GL_FLOAT, image_cube4);
			glTexImage2D(GL_TEXTURE_CUBE_MAP_NEGATIVE_Z_ARB, 0, GL_RGB, image_height/4, image_width/3, 0, GL_RGB, GL_FLOAT, image_cube5);
		} else {
			// TONEMAP
			glTexImage2D(GL_TEXTURE_CUBE_MAP_POSITIVE_X_ARB, 0, GL_RGB, image_height/4, image_width/3, 0, GL_RGB, GL_UNSIGNED_BYTE, image_cube0_tga);
			glTexImage2D(GL_TEXTURE_CUBE_MAP_NEGATIVE_X_ARB, 0, GL_RGB, image_height/4, image_width/3, 0, GL_RGB, GL_UNSIGNED_BYTE, image_cube1_tga);
			glTexImage2D(GL_TEXTURE_CUBE_MAP_POSITIVE_Y_ARB, 0, GL_RGB, image_height/4, image_width/3, 0, GL_RGB, GL_UNSIGNED_BYTE, image_cube2_tga);
			glTexImage2D(GL_TEXTURE_CUBE_MAP_NEGATIVE_Y_ARB, 0, GL_RGB, image_height/4, image_width/3, 0, GL_RGB, GL_UNSIGNED_BYTE, image_cube3_tga);
			glTexImage2D(GL_TEXTURE_CUBE_MAP_POSITIVE_Z_ARB, 0, GL_RGB, image_height/4, image_width/3, 0, GL_RGB, GL_UNSIGNED_BYTE, image_cube4_tga);
			glTexImage2D(GL_TEXTURE_CUBE_MAP_NEGATIVE_Z_ARB, 0, GL_RGB, image_height/4, image_width/3, 0, GL_RGB, GL_UNSIGNED_BYTE, image_cube5_tga);
		}
		bToneMap = !bToneMap;
	}

	if(texFlag==RF_TEX_NOTEXTURE || texFlag==RF_TEX_POINTCUBE)
	{

		if(selectedLightQty>0)
		{

		// =========================================================================================
		//  SCALE 
		// -----------------------------------------------------------------------------------------
		if (c == 43)						//	Upscale Light ADD KEY
		{
			subLight(selectedLight);
			myLights[selectedLight].color[0]*=(float)((myLights[selectedLight].scale*2)-1)*(float)((myLights[selectedLight].scale*2)-1);
			myLights[selectedLight].color[1]*=(float)((myLights[selectedLight].scale*2)-1)*(float)((myLights[selectedLight].scale*2)-1);
			myLights[selectedLight].color[2]*=(float)((myLights[selectedLight].scale*2)-1)*(float)((myLights[selectedLight].scale*2)-1);
			myLights[selectedLight].scale++;
			myLights[selectedLight].color[0]/=((float)((myLights[selectedLight].scale*2)-1)*(float)((myLights[selectedLight].scale*2)-1));
			myLights[selectedLight].color[1]/=((float)((myLights[selectedLight].scale*2)-1)*(float)((myLights[selectedLight].scale*2)-1));
			myLights[selectedLight].color[2]/=((float)((myLights[selectedLight].scale*2)-1)*(float)((myLights[selectedLight].scale*2)-1));
			addLight(selectedLight);
			bChange = true;
		}

		if (c == 45)						// Downscale Light SUBSTRACT KEY
		{
			if (myLights[selectedLight].scale > 1)
			{
				subLight(selectedLight);
				myLights[selectedLight].color[0]*=(float)((myLights[selectedLight].scale*2)-1)*(float)((myLights[selectedLight].scale*2)-1);
				myLights[selectedLight].color[1]*=(float)((myLights[selectedLight].scale*2)-1)*(float)((myLights[selectedLight].scale*2)-1);
				myLights[selectedLight].color[2]*=(float)((myLights[selectedLight].scale*2)-1)*(float)((myLights[selectedLight].scale*2)-1);
				myLights[selectedLight].scale--;
				myLights[selectedLight].color[0]/=((float)((myLights[selectedLight].scale*2)-1)*(float)((myLights[selectedLight].scale*2)-1));
				myLights[selectedLight].color[1]/=((float)((myLights[selectedLight].scale*2)-1)*(float)((myLights[selectedLight].scale*2)-1));
				myLights[selectedLight].color[2]/=((float)((myLights[selectedLight].scale*2)-1)*(float)((myLights[selectedLight].scale*2)-1));
				addLight(selectedLight);
				bChange = true;
			}
		}
		// =========================================================================================


		// =========================================================================================
		//  COLOR CHANGE
		// -----------------------------------------------------------------------------------------
		if (c == 127)						// Decrement Red (DELETE KEY)
		{
			subLight(selectedLight);
			myLights[selectedLight].color[0] -= RF_DELTA_COLOR/((float)((myLights[selectedLight].scale*2)-1)*(float)((myLights[selectedLight].scale*2)-1));

			if (myLights[selectedLight].color[0] < 0.0) myLights[selectedLight].color[0] = 0.0;
			addLight(selectedLight);
			bChange = true;

			printf("Light %.1i: R:%f G:%f B:%f\n",selectedLight, myLights[selectedLight].color[0], myLights[selectedLight].color[1], myLights[selectedLight].color[2]);
		}
		// =========================================================================================
		}

		// =========================================================================================
		//  LIGHT SELECTION / TURN ON/OFF
		// -----------------------------------------------------------------------------------------
		if (c == '0') { checkLight(0); }
		if (c == '1') { checkLight(1); }
		if (c == '2') { checkLight(2); }
		if (c == '3') { checkLight(3); }
		if (c == '4') { checkLight(4); }
		if (c == '5') { checkLight(5); }
		if (c == '6') { checkLight(6); }
		if (c == '7') { checkLight(7); }
		if (c == '8') { checkLight(8); }
		if (c == '9') { checkLight(9); }
		// =========================================================================================
	} 

	// =========================================================================================
	//  OTHER KEYS
	// -----------------------------------------------------------------------------------------
	if (c == 'F' || c == 'f')			// Toggle Mode
	{
		aux = texFlag;
		switch(aux)
		{
		case RF_TEX_NOTEXTURE:
			{
				texFlag = RF_TEX_CUBEMAP;
				bChange = true;
				break;
			}
		case RF_TEX_CUBEMAP:
			{
				texFlag = RF_TEX_POINTCUBE;
				bChange = true;
				break;
			}
		case RF_TEX_POINTCUBE:
			{
				texFlag = RF_TEX_NOTEXTURE;

				for(i = 0; i<(int)RF_BLOCKS; i++)
					s[i][0] = s[i][1] = s[i][2] = 0.0;

				for(i = 0; i<RF_MAXLIGHTS; i++)
					if(myLights[i].on)
						addLight(i);

				bChange = true;
				break;
			}
		}
	}
	// =========================================================================================

	if(texFlag==RF_TEX_CUBEMAP || texFlag==RF_TEX_POINTCUBE)
	{
		if (c == 'Q' || c == 'q')			// Rotate Environment Map
		{
			beginx = 0;		beginy = 0;
			cubex=-5;		cubey=-5;
			W = sw;		H = sh;
			trackball(lastquat, (2.0 * beginx - W) / W, (H - 2.0 * beginy) / H, (2.0 * cubex - W) / W, (H - 2.0 * cubey) / H);
			add_quats(lastquat, curquat, curquat);
	
			bChange = true;
		}

		if (c == 'W' || c == 'w')			// Rotate Environment Map
		{
			beginx = 0;		beginy = 0;
			cubex=0;		cubey=-7;
			W = sw;		H = sh;
			trackball(lastquat, (2.0 * beginx - W) / W, (H - 2.0 * beginy) / H, (2.0 * cubex - W) / W, (H - 2.0 * cubey) / H);
			add_quats(lastquat, curquat, curquat);
	
			bChange = true;
		}

		if (c == 'E' || c == 'e')			// Rotate Environment Map
		{
			beginx = 0;		beginy = 0;
			cubex=5;		cubey=-5;
			W = sw;		H = sh;
			trackball(lastquat, (2.0 * beginx - W) / W, (H - 2.0 * beginy) / H, (2.0 * cubex - W) / W, (H - 2.0 * cubey) / H);
			add_quats(lastquat, curquat, curquat);
	
			bChange = true;
		}

		if (c == 'A' || c == 'a')			// Rotate Environment Map
		{
			beginx = 0;		beginy = 0;
			cubex=-7;		cubey=0;
			W = sw;		H = sh;
			trackball(lastquat, (2.0 * beginx - W) / W, (H - 2.0 * beginy) / H, (2.0 * cubex - W) / W, (H - 2.0 * cubey) / H);
			add_quats(lastquat, curquat, curquat);
	
			bChange = true;
		}

		if (c == 'D' || c == 'd')			// Rotate Environment Map
		{
			beginx = 0;		beginy = 0;
			cubex=7;		cubey=0;
			W = sw;		H = sh;
			trackball(lastquat, (2.0 * beginx - W) / W, (H - 2.0 * beginy) / H, (2.0 * cubex - W) / W, (H - 2.0 * cubey) / H);
			add_quats(lastquat, curquat, curquat);
	
			bChange = true;
		}

		if (c == 'Z' || c == 'z')			// Rotate Environment Map
		{
			beginx = 0;		beginy = 0;
			cubex=-5;		cubey=5;
			W = sw;		H = sh;
			trackball(lastquat, (2.0 * beginx - W) / W, (H - 2.0 * beginy) / H, (2.0 * cubex - W) / W, (H - 2.0 * cubey) / H);
			add_quats(lastquat, curquat, curquat);
	
			bChange = true;
		}

		if (c == 'X' || c == 'x')			// Rotate Environment Map
		{
			beginx = 0;		beginy = 0;
			cubex=0;		cubey=7;
			W = sw;		H = sh;
			trackball(lastquat, (2.0 * beginx - W) / W, (H - 2.0 * beginy) / H, (2.0 * cubex - W) / W, (H - 2.0 * cubey) / H);
			add_quats(lastquat, curquat, curquat);
	
			bChange = true;
		}

		if (c == 'C' || c == 'c')			// Rotate Environment Map
		{
			beginx = 0;		beginy = 0;
			cubex=5;		cubey=5;
			W = sw;		H = sh;
			trackball(lastquat, (2.0 * beginx - W) / W, (H - 2.0 * beginy) / H, (2.0 * cubex - W) / W, (H - 2.0 * cubey) / H);
			add_quats(lastquat, curquat, curquat);
	
			bChange = true;
		}
	}

	glutPostRedisplay();
}
// =========================================================================================


// =========================================================================================
//  Saves a TGA file screenshot
// -----------------------------------------------------------------------------------------
void screenshot(char* filename)
{
	int width, height;

	FILE* fp;
	long i;
	byte header[18];

	width = sw;
	height = sh;
	glReadPixels(0,0,width,height,GL_RGB,GL_UNSIGNED_BYTE,data);

	fp = fopen(filename, "wb");
	header[0] = 0;		// size of ID field that follows 18 byte header (0 usually)
	header[1] = 0;		// type of colour map 0=none, 1=has palette
	header[2] = 2;		// type of image 0=none,1=indexed,2=rgb,3=grey,+8=rle packed
	header[3] = 0;		// first colour map entry in palette
	header[4] = 0;
	header[5] = 0;		// number of colours in palette
	header[6] = 0;
	header[7] = 0;		// number of bits per palette entry 15,16,24,32

	header[8] = 0;		// image x origin
	header[9] = 0;
	header[10] = 0;		// image y origin
	header[11] = 0;
	header[12] = (byte)((width & 0xFF) >> 0);	// image width in pixels low order bits
	header[13] = (byte)((width & 0xFF00) >> 8);	// image width in pixels high order bits
	header[14] = (byte)((height & 0xFF) >> 0);	// image height in pixels low order bits
	header[15] = (byte)((height & 0xFF00) >> 8);	// image height in pixels high order bits
	header[16] = 24;	// image bits per pixel 8,16,24,32
	header[17] = 8;		// image descriptor bits (vh flip bits) I just copied '8' from Photoshop
	//switch R and B from RGB to BGR
	//the coming up loop may also require inverting the image on y
	for (i = 0; i < width*height*3; i += 3) {
		bgraData[i+2] = data[i+0];
		bgraData[i+1] = data[i+1];
		bgraData[i+0] = data[i+2];
	}
	fwrite(header, sizeof(byte), 18, fp);
	fwrite(bgraData, sizeof(byte), width*height*3, fp);
	fclose(fp);
}
// =========================================================================================


// =========================================================================================
//  Special Keys Capturing
// -----------------------------------------------------------------------------------------
void keyboardSpecial(int k, int x, int y)
{
	if (k == GLUT_KEY_F10)		// F10 to toggle Full Screen / Windowed Mode
	{
		if(!bFullScreen)
		{
			oldWidth = sw;
			oldHeight = sh;
			glutFullScreen();
		}
		else
		{
			glutReshapeWindow(oldWidth, oldHeight);	// QUIQUE
		}
		bFullScreen = !bFullScreen;
	}

	if(texFlag==RF_TEX_NOTEXTURE || RF_TEX_POINTCUBE)
	{
		if(selectedLightQty>0)
		{
		// =========================================================================================
		//  LIGHT MOVEMENT
		// -----------------------------------------------------------------------------------------
		if (k == GLUT_KEY_DOWN)						// Move light down
		{
			subLight(selectedLight);
			myLights[selectedLight].yIndex--;
			if(myLights[selectedLight].yIndex==-1) myLights[selectedLight].yIndex+=RF_LIGHTRES;
			addLight(selectedLight);
			bChange = true;
		}

		if (k == GLUT_KEY_UP)						//  Move light up
		{
			subLight(selectedLight);
			myLights[selectedLight].yIndex++;
			if(myLights[selectedLight].yIndex==RF_LIGHTRES) myLights[selectedLight].yIndex-=RF_LIGHTRES;
			addLight(selectedLight);
			bChange = true;
		}

		if (k == GLUT_KEY_LEFT)						// Move light left
		{
			subLight(selectedLight);
			myLights[selectedLight].xIndex--;
			if(myLights[selectedLight].xIndex==-1) myLights[selectedLight].xIndex+=RF_LIGHTRES;
			addLight(selectedLight);
			bChange = true;
		}

		if (k == GLUT_KEY_RIGHT)					// Move light right
		{
			subLight(selectedLight);
			myLights[selectedLight].xIndex++;
			if(myLights[selectedLight].xIndex==RF_LIGHTRES) myLights[selectedLight].xIndex-=RF_LIGHTRES;
			addLight(selectedLight);
			bChange = true;
		}
		// =========================================================================================


		// =========================================================================================
		//  COLOR CHANGE
		// -----------------------------------------------------------------------------------------
		if (k == GLUT_KEY_INSERT)						// Increment Red
		{
			subLight(selectedLight);
			myLights[selectedLight].color[0] += RF_DELTA_COLOR/((float)((myLights[selectedLight].scale*2)-1)*(float)((myLights[selectedLight].scale*2)-1));
			addLight(selectedLight);
			bChange = true;
			printf("Light %.1i: R:%f G:%f B:%f\n",selectedLight, myLights[selectedLight].color[0], myLights[selectedLight].color[1], myLights[selectedLight].color[2]);
		}

		if (k == GLUT_KEY_HOME)							// Increment Green
		{
			subLight(selectedLight);
			myLights[selectedLight].color[1] += RF_DELTA_COLOR/((float)((myLights[selectedLight].scale*2)-1)*(float)((myLights[selectedLight].scale*2)-1));
			addLight(selectedLight);
			bChange = true;

			printf("Light %.1i: R:%f G:%f B:%f\n",selectedLight, myLights[selectedLight].color[0], myLights[selectedLight].color[1], myLights[selectedLight].color[2]);
		}

		if (k == GLUT_KEY_END)							// Decrement Green
		{
			subLight(selectedLight);
			myLights[selectedLight].color[1] -= RF_DELTA_COLOR/((float)((myLights[selectedLight].scale*2)-1)*(float)((myLights[selectedLight].scale*2)-1));
			if (myLights[selectedLight].color[1] < 0.0) myLights[selectedLight].color[1] = 0.0;
			bChange = true;
			addLight(selectedLight);
			printf("Light %.1i: R:%f G:%f B:%f\n",selectedLight, myLights[selectedLight].color[0], myLights[selectedLight].color[1], myLights[selectedLight].color[2]);
		}

		if (k == GLUT_KEY_PAGE_UP)						// Increment Blue
		{
			subLight(selectedLight);
			myLights[selectedLight].color[2] += RF_DELTA_COLOR/((float)((myLights[selectedLight].scale*2)-1)*(float)((myLights[selectedLight].scale*2)-1));
			addLight(selectedLight);
			bChange = true;
			printf("Light %.1i: R:%f G:%f B:%f\n",selectedLight, myLights[selectedLight].color[0], myLights[selectedLight].color[1], myLights[selectedLight].color[2]);
		}

		if (k == GLUT_KEY_PAGE_DOWN)					// Decrement Blue
		{
			subLight(selectedLight);
			myLights[selectedLight].color[2] -= RF_DELTA_COLOR/((float)((myLights[selectedLight].scale*2)-1)*(float)((myLights[selectedLight].scale*2)-1));
			if (myLights[selectedLight].color[2] < 0.0) myLights[selectedLight].color[2] = 0.0;
			addLight(selectedLight);
			bChange = true;
			printf("Light %.1i: R:%f G:%f B:%f\n",selectedLight, myLights[selectedLight].color[0], myLights[selectedLight].color[1], myLights[selectedLight].color[2]);
		}
		// =========================================================================================
		}
	} 

	if (k == GLUT_KEY_F3)					// Divide Exposure by 2
	{
		exposure/=2.0;
		printf("Exposure: %f\n",exposure);
		bChange = true;
	}

	if (k == GLUT_KEY_F4)					// Multiply Exposure by 2
	{
		exposure*=2.0;
		printf("Exposure: %f\n",exposure);
		bChange = true;
	}

	if (k == GLUT_KEY_F2)					// Increment Exposure
	{
		exposure+=0.1;
		printf("Exposure: %f\n",exposure);
		bChange = true;
	}

	if (k == GLUT_KEY_F1)					// Decrement Exposure
	{
		exposure-=0.1;
		printf("Exposure: %f\n",exposure);
		bChange = true;
	}

	if (k == GLUT_KEY_F11)					// Toggle: Generate screenshots from a video file
	{
		loadMatrices = !loadMatrices;
		printf("Generating Screenshots For Video File feature is turned %s\n",loadMatrices ? "ON" : "OFF");
		myVideoFileLoad = fopen(RF_VIDEOFILE,"r");
		bChange = true;
	}

	if (k == GLUT_KEY_F9)					// Toggle: Save rotation matrices in a video file 30FPS
	{
		saveMatrices = !saveMatrices;
		printf("Save Video File feature is turned %s\n",saveMatrices ? "ON" : "OFF");
		lastTime = clock();
		if(!saveMatrices) { if(myVideoFile!=NULL) fclose(myVideoFile); }
	}

	if (k == GLUT_KEY_F5)					// Toggle: Save screenshots in a video file ~30FPS
	{
		saveScreenshot = !saveScreenshot;
		printf("Save Screenshots feature is turned %s\n",saveScreenshot ? "ON" : "OFF");
		lastTime = clock();
	}

	if (k == GLUT_KEY_F7)					// Save screenshot
	{
		char myScreenshot[1024];
		sprintf(myScreenshot, "%s%.8d.tga", RF_VIDEODIR, filenumber);
		screenshot(myScreenshot);
		printf("--> File %s saved.\n",myScreenshot);
		filenumber++;
	}

	glutPostRedisplay();
}
// =========================================================================================


// =========================================================================================
//  Computes Relighting
// -----------------------------------------------------------------------------------------
void generateImage()
{
	
	unsigned int i, j, k, UVsIni;
	int x, y;

	// V * s
	for(i=0;i<RF_RANK_Q;i++)
	{
		Vs[i][0] = 0.0;
		Vs[i][1] = 0.0;
		Vs[i][2] = 0.0;
		for(j=0;j<RF_IMAGES;j++)
		{
			Vs[i][0] += Vr[i][j] * s[j][0];
			Vs[i][1] += Vg[i][j] * s[j][1];
			Vs[i][2] += Vb[i][j] * s[j][2];
		}
	}

	// U*V*s
	for(i=0;i<RF_BLOCKS*RF_RANK_B;i++)
	{
		UVs[i][0] = 0.0;
		UVs[i][1] = 0.0;
		UVs[i][2] = 0.0;
		for(j=0;j<RF_RANK_Q;j++)
		{
			UVs[i][0] += Ur[i][j] * Vs[j][0];
			UVs[i][1] += Ug[i][j] * Vs[j][1];
			UVs[i][2] += Ub[i][j] * Vs[j][2];
		}
	}

	// Ej*U*V*s
	x = 0;
	y = -1;
	int myYIndexAux = - (int)RF_PIXELRES;
	int myXIndexAux = 0;
	UVsIni = 0;

	for(k=0;k<RF_BLOCKS;k++)
	{
		// Matrices Product
		for(i=0;i<RF_PIXELS;i++)
		{
			EjUVs[i][0] = 0.0;
			EjUVs[i][1] = 0.0;
			EjUVs[i][2] = 0.0;
			for(j=0;j<RF_RANK_B;j++)
			{
				EjUVs[i][0] += Er[k][i][j] * UVs[UVsIni+j][0];
				EjUVs[i][1] += Eg[k][i][j] * UVs[UVsIni+j][1];
				EjUVs[i][2] += Eb[k][i][j] * UVs[UVsIni+j][2];
			}
		}
		// Final Image Block
		y++;
		myYIndexAux+=RF_PIXELRES;
		if (y==(int)RF_BLOCKRESW)
		{
			x++;
			myXIndexAux += RF_PIXELRES;
			y = 0;
			myYIndexAux = 0;
		}
		unsigned int myIndexAux = 0;
		for(i=0;i<RF_PIXELRES;i++)
		{
			int myXIndex = RF_IMAGERESH - 1 - myXIndexAux - ((int)i);
			for(j=0;j<RF_PIXELRES;j++)
			{
				int myYIndex = myYIndexAux + j;
				unsigned int myIndex = myIndexAux + j;
				unsigned long myIIndex = indexI(myXIndex,myYIndex,0);
				I[myIIndex] = (float) EjUVs[myIndex][0];
				I[myIIndex+1] = (float) EjUVs[myIndex][1];
				I[myIIndex+2] = (float) EjUVs[myIndex][2];
			}
			myIndexAux += RF_PIXELRES;
		}
		UVsIni+=RF_RANK_B;
	}
}
// =========================================================================================


// =========================================================================================
//  Draw Relighted Image
// -----------------------------------------------------------------------------------------
void drawImage()
{
	glRasterPos2f(5.0f*(1.0f-(float)(RF_IMAGERESW)/(float)(sw)),50.0f*(1.0f-(float)(RF_IMAGERESH)/(float)(sh)));
	glDrawPixels(RF_IMAGERESW, RF_IMAGERESH, GL_RGB, GL_FLOAT, I);
}
// =========================================================================================


// =========================================================================================
//  Draw CubeMap Cross
// -----------------------------------------------------------------------------------------
void drawCruxContent(void)
{
	float xc, yc, delta;

	xc = RF_CRUX_XC;
	yc = RF_CRUX_YC;
	delta = RF_CRUX_DELTA;

	glEnable(GL_TEXTURE_CUBE_MAP_ARB);
	glEnable(GL_DEPTH_TEST);
    glBegin(GL_QUADS);
		// FRONT 
		glTexCoord3f(-1,-1,1);
		glVertex3f(xc - delta/2, yc, 0);
		glTexCoord3f(-1,1,1);
		glVertex3f(xc - delta/2, yc + delta, 0);
		glTexCoord3f(1,1,1);
		glVertex3f(xc + delta/2, yc + delta, 0);
		glTexCoord3f(1,-1,1);
		glVertex3f(xc + delta/2, yc, 0);

		// BACK
		glTexCoord3f(1,-1,-1);
		glVertex3f(xc + 3*delta/2, yc, 0);
		glTexCoord3f(1,1,-1);
		glVertex3f(xc + 3*delta/2, yc + delta, 0);
		glTexCoord3f(-1,1,-1);
		glVertex3f(xc + 5*delta/2, yc + delta, 0);
		glTexCoord3f(-1,-1,-1);
		glVertex3f(xc + 5*delta/2, yc, 0); 

		// TOP
		glTexCoord3f(-1,1,1);
		glVertex3f(xc - delta/2, yc + delta, 0);
		glTexCoord3f(-1,1,-1);
		glVertex3f(xc - delta/2, yc + 2*delta, 0);
		glTexCoord3f(1,1,-1);
		glVertex3f(xc + delta/2, yc + 2*delta, 0);
		glTexCoord3f(1,1,1);
		glVertex3f(xc + delta/2, yc + delta, 0); 

		// BOTTOM
		glTexCoord3f(-1,-1,-1);
		glVertex3f(xc - delta/2, yc - delta, 0);
		glTexCoord3f(-1,-1,1);
		glVertex3f(xc - delta/2, yc, 0);
		glTexCoord3f(1,-1,1);
		glVertex3f(xc + delta/2, yc, 0);
		glTexCoord3f(1,-1,-1);
		glVertex3f(xc + delta/2, yc - delta, 0); 

		// LEFT
		glTexCoord3f(-1,-1,-1);
		glVertex3f(xc - 3*delta/2, yc, 0);
		glTexCoord3f(-1,1,-1);
		glVertex3f(xc - 3*delta/2, yc + delta, 0);
		glTexCoord3f(-1,1,1);
		glVertex3f(xc - delta/2, yc + delta, 0);
		glTexCoord3f(-1,-1,1);
		glVertex3f(xc - delta/2, yc, 0); 

		// RIGHT
		glTexCoord3f(1,-1,1);
		glVertex3f(xc + delta/2, yc, 0);
		glTexCoord3f(1,1,1);
		glVertex3f(xc + delta/2, yc + delta, 0);
		glTexCoord3f(1,1,-1);
		glVertex3f(xc + 3*delta/2, yc + delta, 0);
		glTexCoord3f(1,-1,-1);
		glVertex3f(xc + 3*delta/2, yc, 0); 
	glEnd();

	glDisable(GL_TEXTURE_CUBE_MAP_ARB);
	glDisable(GL_DEPTH_TEST);
}

void drawCruxOutline(void)
{
	float xc, yc, delta;

	xc = RF_CRUX_XC;
	yc = RF_CRUX_YC;
	delta = RF_CRUX_DELTA;

	glDisable(GL_TEXTURE_CUBE_MAP_ARB);
	glDisable(GL_DEPTH_TEST);
	glColor4f(1,1,1,1);
    glBegin(GL_LINE_LOOP);
		glVertex3f(xc - delta/2, yc, 0);	// Front
		glVertex3f(xc - delta/2, yc + delta, 0);
		glVertex3f(xc + delta/2, yc + delta, 0);
		glVertex3f(xc + delta/2, yc, 0); 
	glEnd();
    glBegin(GL_LINE_LOOP);
		glVertex3f(xc - delta/2, yc + delta, 0);	// Up
		glVertex3f(xc - delta/2, yc + 2*delta, 0);
		glVertex3f(xc + delta/2, yc + 2*delta, 0);
		glVertex3f(xc + delta/2, yc + delta, 0); 
	glEnd();
    glBegin(GL_LINE_LOOP);
		glVertex3f(xc - delta/2, yc - delta, 0);	// Bottom
		glVertex3f(xc - delta/2, yc, 0);
		glVertex3f(xc + delta/2, yc, 0);
		glVertex3f(xc + delta/2, yc - delta, 0); 
	glEnd();
    glBegin(GL_LINE_LOOP);
		glVertex3f(xc + 3*delta/2, yc, 0);	// Back
		glVertex3f(xc + 3*delta/2, yc + delta, 0);
		glVertex3f(xc + 5*delta/2, yc + delta, 0);
		glVertex3f(xc + 5*delta/2, yc, 0); 
	glEnd();
    glBegin(GL_LINE_LOOP);
		glVertex3f(xc - 3*delta/2, yc, 0);	// Left
		glVertex3f(xc - 3*delta/2, yc + delta, 0);
		glVertex3f(xc - delta/2, yc + delta, 0);
		glVertex3f(xc - delta/2, yc, 0); 
	glEnd();
    glBegin(GL_LINE_LOOP);
		glVertex3f(xc + delta/2, yc, 0);	// Right
		glVertex3f(xc + delta/2, yc + delta, 0);
		glVertex3f(xc + 3*delta/2, yc + delta, 0);
		glVertex3f(xc + 3*delta/2, yc, 0); 
	glEnd();
	glEnable(GL_TEXTURE_CUBE_MAP_ARB);
	glEnable(GL_DEPTH_TEST);
}

void drawEnvironment(void)
{
	float min[3];
	float max[3];

	float center = 5.0f*(1.0f-(float)(RF_IMAGERESW)/(float)(sw)) + 50.0f*((float)(RF_IMAGERESW)/(float)(sw));

	min[0] =  center-50.0f;
	min[1] =  0.0f;
	min[2] =  0.0f;

	max[0] =  center+50.0f;
	max[1] =  100.0f;
	max[2] =  100.0f;

	glEnable(GL_TEXTURE_CUBE_MAP_ARB);
	glDepthMask(GL_FALSE);
	glDisable(GL_DEPTH_TEST);

    glBegin(GL_QUADS);
		// FRONT 
		glTexCoord3f(-1,-1,1);
		glVertex3f(min[0], min[1], min[2]);
		glTexCoord3f(-1,1,1);
		glVertex3f(min[0], max[1], min[2]);
		glTexCoord3f(1,1,1);
		glVertex3f(max[0], max[1], min[2]);
		glTexCoord3f(1,-1,1);
		glVertex3f(max[0], min[1], min[2]);

		// BACK
		glTexCoord3f(-1,1,-1);
		glVertex3f(min[0], min[1], max[2]);
		glTexCoord3f(-1,-1,-1);
		glVertex3f(min[0], max[1], max[2]);
		glTexCoord3f(1,-1,-1);
		glVertex3f(max[0], max[1], max[2]);
		glTexCoord3f(1,1,-1);
		glVertex3f(max[0], min[1], max[2]); 

		// TOP
		glTexCoord3f(-1,1,1);
		glVertex3f(min[0], max[1], min[2]);
		glTexCoord3f(-1,1,-1);
		glVertex3f(min[0], max[1], max[2]);
		glTexCoord3f(1,1,-1);
		glVertex3f(max[0], max[1], max[2]);
		glTexCoord3f(1,1,1);
		glVertex3f(max[0], max[1], min[2]); 

		// BOTTOM
		glTexCoord3f(-1,-1,-1);
		glVertex3f(min[0], min[1], min[2]);
		glTexCoord3f(-1,-1,1);
		glVertex3f(min[0], min[1], max[2]);
		glTexCoord3f(1,-1,1);
		glVertex3f(max[0], min[1], max[2]);
		glTexCoord3f(1,-1,-1);
		glVertex3f(max[0], min[1], min[2]); 

		// LEFT
		glTexCoord3f(-1,-1,-1);
		glVertex3f(min[0], min[1], min[2]);
		glTexCoord3f(-1,1,-1);
		glVertex3f(min[0], max[1], min[2]);
		glTexCoord3f(-1,1,1);
		glVertex3f(min[0], max[1], max[2]);
		glTexCoord3f(-1,-1,1);
		glVertex3f(min[0], min[1], max[2]); 

		// RIGHT
		glTexCoord3f(1,-1,1);
		glVertex3f(max[0], min[1], min[2]);
		glTexCoord3f(1,1,1);
		glVertex3f(max[0], max[1], min[2]);
		glTexCoord3f(1,1,-1);
		glVertex3f(max[0], max[1], max[2]);
		glTexCoord3f(1,-1,-1);
		glVertex3f(max[0], min[1], max[2]); 
	glEnd();

	glDisable(GL_TEXTURE_CUBE_MAP_ARB);
}
// =========================================================================================


// =========================================================================================
//  Draw Light Status Panels
// -----------------------------------------------------------------------------------------
void drawLightStatus(void)
{
	float x0, y0, scale;

	x0 = RF_LIGHTSPANEL_X0;
	y0 = RF_LIGHTSPANEL_Y0;
	scale = RF_LIGHTSPANEL_SCALE * (100.0f / (float)sh);

	glDisable(GL_DEPTH_TEST);
	glColor4f(1,1,1,1);

	for(int i=0; i<10; i++)
	{
		if(selectedLight==i)
			glColor3f(1, 1, 0);
		else
			glColor3f(1, 1, 1);

		// L
		glBegin(GL_LINE_STRIP);
			glVertex3f(x0, y0+4*scale, 0);
			glVertex3f(x0, y0, 0);
			glVertex3f(x0+2*scale, y0, 0);
		glEnd();

		switch(i)
		{
		case 0:
			glBegin(GL_LINE_LOOP);
				glVertex3f(x0+4*scale, y0+4*scale, 0);
				glVertex3f(x0+7*scale, y0+4*scale, 0);
				glVertex3f(x0+7*scale, y0, 0);
				glVertex3f(x0+4*scale, y0, 0);
			glEnd();
			break;
		case 1:
			glBegin(GL_LINE_STRIP);
				glVertex3f(x0+4*scale, y0+4*scale, 0);
				glVertex3f(x0+6*scale, y0+4*scale, 0);
				glVertex3f(x0+6*scale, y0, 0);
			glEnd();
			break;
		case 2:
			glBegin(GL_LINE_STRIP);
				glVertex3f(x0+4*scale, y0+4*scale, 0);
				glVertex3f(x0+7*scale, y0+4*scale, 0);
				glVertex3f(x0+7*scale, y0+2*scale, 0);
				glVertex3f(x0+4*scale, y0+2*scale, 0);
				glVertex3f(x0+4*scale, y0, 0);
				glVertex3f(x0+7*scale, y0, 0);
			glEnd();
			break;
		case 3:
			glBegin(GL_LINES);
				glVertex3f(x0+4*scale, y0+4*scale, 0);
				glVertex3f(x0+7*scale, y0+4*scale, 0);
				glVertex3f(x0+7*scale, y0+4*scale, 0);
				glVertex3f(x0+7*scale, y0, 0);
				glVertex3f(x0+7*scale, y0, 0);
				glVertex3f(x0+4*scale, y0, 0);
				glVertex3f(x0+6*scale, y0+2*scale, 0);
				glVertex3f(x0+7*scale, y0+2*scale, 0);
			glEnd();
			break;
		case 4:
			glBegin(GL_LINE_STRIP);
				glVertex3f(x0+4*scale, y0+4*scale, 0);
				glVertex3f(x0+4*scale, y0+2*scale, 0);
				glVertex3f(x0+7*scale, y0+2*scale, 0);
			glEnd();
			glBegin(GL_LINES);
				glVertex3f(x0+7*scale, y0+4*scale, 0);
				glVertex3f(x0+7*scale, y0, 0);
			glEnd();
			break;
		case 5:
			glBegin(GL_LINE_STRIP);
				glVertex3f(x0+7*scale, y0+4*scale, 0);
				glVertex3f(x0+4*scale, y0+4*scale, 0);
				glVertex3f(x0+4*scale, y0+2*scale, 0);
				glVertex3f(x0+7*scale, y0+2*scale, 0);
				glVertex3f(x0+7*scale, y0, 0);
				glVertex3f(x0+4*scale, y0, 0);
			glEnd();
			break;
		case 6:
			glBegin(GL_LINE_STRIP);
				glVertex3f(x0+7*scale, y0+4*scale, 0);
				glVertex3f(x0+4*scale, y0+4*scale, 0);
				glVertex3f(x0+4*scale, y0, 0);
				glVertex3f(x0+7*scale, y0, 0);
				glVertex3f(x0+7*scale, y0+2*scale, 0);
				glVertex3f(x0+4*scale, y0+2*scale, 0);
			glEnd();
			break;
		case 7:
			glBegin(GL_LINE_STRIP);
				glVertex3f(x0+4*scale, y0+4*scale, 0);
				glVertex3f(x0+7*scale, y0+4*scale, 0);
				glVertex3f(x0+6*scale, y0, 0);
			glEnd();
			break;
		case 8:
			glBegin(GL_LINE_STRIP);
				glVertex3f(x0+7*scale, y0+2*scale, 0);
				glVertex3f(x0+7*scale, y0+4*scale, 0);
				glVertex3f(x0+4*scale, y0+4*scale, 0);
				glVertex3f(x0+4*scale, y0+2*scale, 0);
				glVertex3f(x0+7*scale, y0+2*scale, 0);
				glVertex3f(x0+7*scale, y0, 0);
				glVertex3f(x0+4*scale, y0, 0);
				glVertex3f(x0+4*scale, y0+2*scale, 0);
			glEnd();
			break;
		case 9:
			glBegin(GL_LINE_STRIP);
				glVertex3f(x0+7*scale, y0, 0);
				glVertex3f(x0+7*scale, y0+4*scale, 0);
				glVertex3f(x0+4*scale, y0+4*scale, 0);
				glVertex3f(x0+4*scale, y0+2*scale, 0);
				glVertex3f(x0+7*scale, y0+2*scale, 0);
			glEnd();
			break;
		}

		// On / Off Status
		if(myLights[i].on)
			glColor3f(0, 0.8, 0);
		else
			glColor3f(0.8, 0, 0);
		glBegin(GL_QUADS);
			glVertex3f(x0+10*scale, y0, 0);
			glVertex3f(x0+10*scale, y0+4*scale, 0);
			glVertex3f(x0+14*scale, y0+4*scale, 0);
			glVertex3f(x0+14*scale, y0, 0);
		glEnd();
		glColor3f(1, 1, 1);
		glBegin(GL_LINE_LOOP);
			glVertex3f(x0+10*scale, y0, 0);
			glVertex3f(x0+10*scale, y0+4*scale, 0);
			glVertex3f(x0+14*scale, y0+4*scale, 0);
			glVertex3f(x0+14*scale, y0, 0);
		glEnd();

		
		// Light Color
		glColor3f(myLights[i].color[0]*(float)((myLights[i].scale*2)-1)*(float)((myLights[i].scale*2)-1), myLights[i].color[1]*(float)((myLights[i].scale*2)-1)*(float)((myLights[i].scale*2)-1), myLights[i].color[2]*(float)((myLights[i].scale*2)-1)*(float)((myLights[i].scale*2)-1));
		glBegin(GL_QUADS);
			glVertex3f(x0+17*scale, y0+1*scale, 0);
			glVertex3f(x0+17*scale, y0+3*scale, 0);
			glVertex3f(x0+23*scale, y0+3*scale, 0);
			glVertex3f(x0+23*scale, y0+1*scale, 0);
		glEnd();

		// Color Change Keys
		if(selectedLight == i)
		{
			glColor3f(1, 0, 0);
			glBegin(GL_TRIANGLES);
				glVertex3f(x0+26*scale, y0+3*scale, 0);
				glVertex3f(x0+29*scale, y0+7*scale, 0);
				glVertex3f(x0+32*scale, y0+3*scale, 0);
				glVertex3f(x0+26*scale, y0+1*scale, 0);
				glVertex3f(x0+29*scale, y0-3*scale, 0);
				glVertex3f(x0+32*scale, y0+1*scale, 0);
			glEnd();
			glColor3f(0, 1, 0);
			glBegin(GL_TRIANGLES);
				glVertex3f(x0+34*scale, y0+3*scale, 0);
				glVertex3f(x0+37*scale, y0+7*scale, 0);
				glVertex3f(x0+40*scale, y0+3*scale, 0);
				glVertex3f(x0+34*scale, y0+1*scale, 0);
				glVertex3f(x0+37*scale, y0-3*scale, 0);
				glVertex3f(x0+40*scale, y0+1*scale, 0);
			glEnd();
			glColor3f(0, 0, 1);
			glBegin(GL_TRIANGLES);
				glVertex3f(x0+42*scale, y0+3*scale, 0);
				glVertex3f(x0+45*scale, y0+7*scale, 0);
				glVertex3f(x0+48*scale, y0+3*scale, 0);
				glVertex3f(x0+42*scale, y0+1*scale, 0);
				glVertex3f(x0+45*scale, y0-3*scale, 0);
				glVertex3f(x0+48*scale, y0+1*scale, 0);
			glEnd();

			glColor3f(.75, .75, .75);
			glBegin(GL_LINE_LOOP);
				glVertex3f(x0+26*scale, y0+3*scale, 0);
				glVertex3f(x0+29*scale, y0+7*scale, 0);
				glVertex3f(x0+32*scale, y0+3*scale, 0);
			glEnd();
			glBegin(GL_LINE_LOOP);
				glVertex3f(x0+26*scale, y0+1*scale, 0);
				glVertex3f(x0+29*scale, y0-3*scale, 0);
				glVertex3f(x0+32*scale, y0+1*scale, 0);
			glEnd();
			glBegin(GL_LINE_LOOP);
				glVertex3f(x0+34*scale, y0+3*scale, 0);
				glVertex3f(x0+37*scale, y0+7*scale, 0);
				glVertex3f(x0+40*scale, y0+3*scale, 0);
			glEnd();
			glBegin(GL_LINE_LOOP);
				glVertex3f(x0+34*scale, y0+1*scale, 0);
				glVertex3f(x0+37*scale, y0-3*scale, 0);
				glVertex3f(x0+40*scale, y0+1*scale, 0);
			glEnd();
			glBegin(GL_LINE_LOOP);
				glVertex3f(x0+42*scale, y0+3*scale, 0);
				glVertex3f(x0+45*scale, y0+7*scale, 0);
				glVertex3f(x0+48*scale, y0+3*scale, 0);
			glEnd();
			glBegin(GL_LINE_LOOP);
				glVertex3f(x0+42*scale, y0+1*scale, 0);
				glVertex3f(x0+45*scale, y0-3*scale, 0);
				glVertex3f(x0+48*scale, y0+1*scale, 0);
			glEnd();
		}
		
		y0 -= RF_LIGHTSPANEL_YSTEP;
	}

	glEnable(GL_DEPTH_TEST);
}
// =========================================================================================


// =========================================================================================
//  Draw Loop
// -----------------------------------------------------------------------------------------
void Draw(void)
{
	unsigned int i;
	GLfloat m[4][4];											// For CubeMap Rotation

	glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);		// Clear Screen And Depth Buffer
	glLoadIdentity ();											// Reset The Modelview Matrix

	// CUBEMAP ROTATION
	if(texFlag==RF_TEX_CUBEMAP || texFlag==RF_TEX_POINTCUBE)
	{
		glMatrixMode(GL_TEXTURE);
		glPushMatrix();

		// This is to use rotation matrices from the video file
		if(loadMatrices && myVideoFileLoad!=NULL)
		{
			m[3][0] = m[3][1] = m[3][2] = m[0][3] = m[1][3] = m[2][3] = 0.0f;
			m[3][3] = 1.0f;
			int items = fscanf(myVideoFileLoad, "%f %f %f %f %f %f %f %f %f\n",&matrix[0][0],&matrix[0][1],&matrix[0][2],&matrix[1][0],&matrix[1][1],&matrix[1][2],&matrix[2][0],&matrix[2][1],&matrix[2][2]);
			if (items!=9)
			{
				loadMatrices = false;
				fclose(myVideoFileLoad);
				printf("All Screenshots Were Generated!!!\n");
			}
		}

		// This is to use current rotation matrix
		if(!loadMatrices)
		{
			build_rotmatrix(m, curquat);

			for(int iiii=0;iiii<4;iiii++)
				for(int jjjj=0;jjjj<4;jjjj++)
					matrix[iiii][jjjj] = m[jjjj][iiii];
		}

		glMultMatrixf(&matrix[0][0]);

		// Apply rotation and get corresponding texels
		multAndGetTexelCubeMapDirections(myCMDirs, myRotCMDirs, matrix, false, s, image_cube0, image_cube1, image_cube2, image_cube3, image_cube4, image_cube5, image_width/3); // SE_CUBEMAP

		glMatrixMode(GL_MODELVIEW);

		// Add point lights to the s lighting vector
		if(texFlag==RF_TEX_POINTCUBE)
		{
			for(i = 0; i<RF_MAXLIGHTS; i++)
				if(myLights[i].on)
					addLight(i);
		}
	}

	// Image generation (if needed) and draw
	if(bChange)
	{
		generateImage();
		bChange = false;
	}
	drawImage();

	// Draw EnvMap on Screen if needed (cross)
	switch(texFlag)
	{
	case RF_TEX_NOTEXTURE:	break;
	case RF_TEX_CUBEMAP:
	case RF_TEX_POINTCUBE:
		glEnable(GL_TEXTURE_CUBE_MAP_ARB);
		drawCruxContent();					// drawed 2nd time to avoid superposition
		drawCruxOutline();
		glDisable(GL_TEXTURE_CUBE_MAP_ARB);
		break;
	}

	// CUBEMAP ROTATION RESTORE
	if(texFlag==RF_TEX_CUBEMAP || texFlag==RF_TEX_POINTCUBE)
	{
		glMatrixMode(GL_TEXTURE); //
		glPopMatrix();
		glMatrixMode(GL_MODELVIEW); //
	}

	int left = RF_LIGHTSPANEL_X0 - RF_LIGHTRES - 2;
	int bottom = RF_LIGHTSPANEL_Y0 - RF_LIGHTRES;

	if(texFlag==RF_TEX_NOTEXTURE || texFlag==RF_TEX_POINTCUBE)
	{
		// Draw lights
		glBegin(GL_QUADS);
			for(i=0;i<RF_MAXLIGHTS;i++)
			{
				if(myLights[i].on)
				{
					glColor3f(myLights[i].color[0]*(float)((myLights[i].scale*2)-1)*(float)((myLights[i].scale*2)-1), myLights[i].color[1]*(float)((myLights[i].scale*2)-1)*(float)((myLights[i].scale*2)-1), myLights[i].color[2]*(float)((myLights[i].scale*2)-1)*(float)((myLights[i].scale*2)-1));
					glVertex2i(left+myLights[i].xIndex-myLights[i].scale+1, bottom+myLights[i].yIndex-myLights[i].scale+1);
					glVertex2i(left+myLights[i].xIndex-myLights[i].scale+1, bottom+myLights[i].yIndex+myLights[i].scale);
					glVertex2i(left+myLights[i].xIndex+myLights[i].scale, bottom+myLights[i].yIndex+myLights[i].scale);
					glVertex2i(left+myLights[i].xIndex+myLights[i].scale, bottom+myLights[i].yIndex-myLights[i].scale+1);
				}
			}
		glEnd();
	}

	if(texFlag==RF_TEX_NOTEXTURE || texFlag==RF_TEX_POINTCUBE)
	{
		// Draw grid
		glColor4f(0.8f, 0.8f, 0.8f, 0.8f);
		glBegin(GL_LINES);
		for(i=0;i<=RF_LIGHTRES;i++)
		{
			glVertex2i(left, bottom+i);
			glVertex2i(left+RF_LIGHTRES, bottom+i);
			glVertex2i(left+i, bottom);
			glVertex2i(left+i, bottom+RF_LIGHTRES);
		}
		glEnd();
	}

	if(texFlag==RF_TEX_NOTEXTURE || texFlag==RF_TEX_POINTCUBE)
	{
		// Highlight on the grid selected light
		if(selectedLightQty>0)
		{
		glColor3f(0.0f, 1.0f, 0.0f);
		glLineWidth(2.0);
		glBegin(GL_LINE_LOOP);
			glVertex2i(left+myLights[selectedLight].xIndex-myLights[selectedLight].scale+1, bottom+myLights[selectedLight].yIndex-myLights[selectedLight].scale+1);
			glVertex2i(left+myLights[selectedLight].xIndex-myLights[selectedLight].scale+1, bottom+myLights[selectedLight].yIndex+myLights[selectedLight].scale);
			glVertex2i(left+myLights[selectedLight].xIndex+myLights[selectedLight].scale, bottom+myLights[selectedLight].yIndex+myLights[selectedLight].scale);
			glVertex2i(left+myLights[selectedLight].xIndex+myLights[selectedLight].scale, bottom+myLights[selectedLight].yIndex-myLights[selectedLight].scale+1);
		glEnd();
		glLineWidth(1.0);
		}

		drawLightStatus();
	}

	glFlush ();													// Flush The GL Rendering Pipeline
}
// =========================================================================================


// =========================================================================================
//  Reshape Callback Function
// -----------------------------------------------------------------------------------------
void reshape(int w, int h)
{
	sw = w;
	sh = h;

	glViewport(0, 0, w, h);
	glMatrixMode (GL_PROJECTION);										// Select The Projection Matrix
	glLoadIdentity ();													// Reset The Projection Matrix
	gluOrtho2D(0.0f, 100.0f, 0.0, 100.0f);
	glMatrixMode (GL_MODELVIEW);										// Select The Modelview Matrix
	glLoadIdentity ();													// Reset The Modelview Matrix

	lightPanelPrecomputation();
}
// =========================================================================================


// =========================================================================================
//  Display Callback Function
// -----------------------------------------------------------------------------------------
void display(void)
{
	// To generate video file (rotation matrices) if flag active
	if(saveMatrices)
	{
		currentTime = clock();
		elapsedTime = (currentTime - lastTime);
		if(elapsedTime >= RF_FREQUENCY)
		{
			lastTime = currentTime - (elapsedTime - RF_FREQUENCY);
			genVideo();
		}
	}

	Draw();
	glutSwapBuffers();

	// To save images if loading rotation matrices from video file
	if(loadMatrices)
	{
		char myScreenshot[1024];
		sprintf(myScreenshot, "%s%.8d.tga", RF_VIDEODIR, filenumber);
		screenshot(myScreenshot);
		printf("--> File %s saved.\n",myScreenshot);
		filenumber++;
		bChange = true;
	}

	// To generate screenshots if flag active
	if(saveScreenshot)
	{
		currentTime = clock();
		elapsedTime = (currentTime - lastTime);
		if(elapsedTime >= RF_FREQUENCY)
		{
			lastTime = currentTime - (elapsedTime - RF_FREQUENCY);
			char myScreenshot[1024];
			sprintf(myScreenshot, "%s%.8d.tga", RF_VIDEODIR, filenumber);
			screenshot(myScreenshot);
			printf("--> File %s saved.\n",myScreenshot);
			filenumber++;
			bChange = true;
		}
	}

}
// =========================================================================================


// =========================================================================================
//  Menu
// -----------------------------------------------------------------------------------------
void menu(int item)
{
	int i;

	switch(item)
	{
	case RF_TEX_NOTEXTURE:
		{
			texFlag = RF_TEX_NOTEXTURE;

			for(i = 0; i<(int)RF_IMAGES; i++)
				s[i][0] = s[i][1] = s[i][2] = 0.0;

			for(i = 0; i<RF_MAXLIGHTS; i++)
				if(myLights[i].on)
					addLight(i);

			bChange = true;
			break;
		}
	case RF_TEX_CUBEMAP:
		{
			for(i = 0; i<(int)RF_IMAGES; i++)
				s[i][0] = s[i][1] = s[i][2] = 0.0;

			texFlag = RF_TEX_CUBEMAP;
			bChange = true;
			break;
		}
	case RF_TEX_POINTCUBE:
		{
			texFlag = RF_TEX_POINTCUBE;
			bChange = true;
			break;
		}
	}
	glutPostRedisplay();
}
// =========================================================================================


// =========================================================================================
//  Mouse
// -----------------------------------------------------------------------------------------
void mouse(int button, int state, int x, int y)
{
	// To generate video file (rotation matrices) if flag active
	if(saveMatrices)
	{
		currentTime = clock();
		elapsedTime = (currentTime - lastTime);
		if(elapsedTime >= RF_FREQUENCY)
		{
			lastTime = currentTime - (elapsedTime - RF_FREQUENCY);
			genVideo();
		}
	}

	// To generate screenshots if flag active
	if(saveScreenshot)
	{
		currentTime = clock();
		elapsedTime = (currentTime - lastTime);
		if(elapsedTime >= RF_FREQUENCY)
		{
			lastTime = currentTime - (elapsedTime - RF_FREQUENCY);
			char myScreenshot[1024];
			sprintf(myScreenshot, "%s%.8d.tga", RF_VIDEODIR, filenumber);
			screenshot(myScreenshot);
			printf("--> File %s saved.\n",myScreenshot);
			filenumber++;
			bChange = true;
		}
	}

	// Check for button pressed in lights panel
	if(texFlag == RF_TEX_NOTEXTURE || texFlag == RF_TEX_POINTCUBE)
	{
		if (button == GLUT_LEFT_BUTTON && state == GLUT_DOWN)
		{
			float scale = RF_LIGHTSPANEL_SCALE * (100.0f / (float)sh);

			if(x >= lightPanelSel[0][0] && x <= lightPanelSel[0][1] && y >= lightPanelSel[0][2] && y <= lightPanelSel[0][3]) {keyboard(lightPanelKey[0], x, y); return;}
			if(x >= lightPanelSel[1][0] && x <= lightPanelSel[1][1] && y >= lightPanelSel[1][2] && y <= lightPanelSel[1][3]) {keyboard(lightPanelKey[1], x, y); return;}
			if(x >= lightPanelSel[2][0] && x <= lightPanelSel[2][1] && y >= lightPanelSel[2][2] && y <= lightPanelSel[2][3]) {keyboard(lightPanelKey[2], x, y); return;}
			if(x >= lightPanelSel[3][0] && x <= lightPanelSel[3][1] && y >= lightPanelSel[3][2] && y <= lightPanelSel[3][3]) {keyboard(lightPanelKey[3], x, y); return;}
			if(x >= lightPanelSel[4][0] && x <= lightPanelSel[4][1] && y >= lightPanelSel[4][2] && y <= lightPanelSel[4][3]) {keyboard(lightPanelKey[4], x, y); return;}
			if(x >= lightPanelSel[5][0] && x <= lightPanelSel[5][1] && y >= lightPanelSel[5][2] && y <= lightPanelSel[5][3]) {keyboard(lightPanelKey[5], x, y); return;}
			if(x >= lightPanelSel[6][0] && x <= lightPanelSel[6][1] && y >= lightPanelSel[6][2] && y <= lightPanelSel[6][3]) {keyboard(lightPanelKey[6], x, y); return;}
			if(x >= lightPanelSel[7][0] && x <= lightPanelSel[7][1] && y >= lightPanelSel[7][2] && y <= lightPanelSel[7][3]) {keyboard(lightPanelKey[7], x, y); return;}
			if(x >= lightPanelSel[8][0] && x <= lightPanelSel[8][1] && y >= lightPanelSel[8][2] && y <= lightPanelSel[8][3]) {keyboard(lightPanelKey[8], x, y); return;}
			if(x >= lightPanelSel[9][0] && x <= lightPanelSel[9][1] && y >= lightPanelSel[9][2] && y <= lightPanelSel[9][3]) {keyboard(lightPanelKey[9], x, y); return;}

			if(selectedLightQty>0)
			{
				if(x >= lightPanelColor[selectedLight][0][0] && x <= lightPanelColor[selectedLight][0][1] && y >= lightPanelColor[selectedLight][0][2] && y <= lightPanelColor[selectedLight][0][3]) {keyboardSpecial(GLUT_KEY_INSERT, x, y); changingcolor = -1; return;}
				if(x >= lightPanelColor[selectedLight][1][0] && x <= lightPanelColor[selectedLight][1][1] && y >= lightPanelColor[selectedLight][1][2] && y <= lightPanelColor[selectedLight][1][3]) {keyboard(127, x, y); changingcolor = -1; return;}
				if(x >= lightPanelColor[selectedLight][2][0] && x <= lightPanelColor[selectedLight][2][1] && y >= lightPanelColor[selectedLight][2][2] && y <= lightPanelColor[selectedLight][2][3]) {keyboardSpecial(GLUT_KEY_HOME, x, y); changingcolor = -1; return;}
				if(x >= lightPanelColor[selectedLight][3][0] && x <= lightPanelColor[selectedLight][3][1] && y >= lightPanelColor[selectedLight][3][2] && y <= lightPanelColor[selectedLight][3][3]) {keyboardSpecial(GLUT_KEY_END, x, y); changingcolor = -1; return;}
				if(x >= lightPanelColor[selectedLight][4][0] && x <= lightPanelColor[selectedLight][4][1] && y >= lightPanelColor[selectedLight][4][2] && y <= lightPanelColor[selectedLight][4][3]) {keyboardSpecial(GLUT_KEY_PAGE_UP, x, y); changingcolor = -1; return;}
				if(x >= lightPanelColor[selectedLight][5][0] && x <= lightPanelColor[selectedLight][5][1] && y >= lightPanelColor[selectedLight][5][2] && y <= lightPanelColor[selectedLight][5][3]) {keyboardSpecial(GLUT_KEY_PAGE_DOWN, x, y); changingcolor = -1; return;}
			}
		}
	}


	// Check for button pressed in lights grid
	int left = (RF_LIGHTSPANEL_X0 - RF_LIGHTRES - 2);
	int bottom = (RF_LIGHTSPANEL_Y0 - RF_LIGHTRES);

	if(texFlag == RF_TEX_NOTEXTURE || texFlag == RF_TEX_POINTCUBE)
	{
		if (button == GLUT_LEFT_BUTTON && state == GLUT_DOWN)
		{
			if(x >= (int)((left * sw) / 100.0f) && x <= (int)(((left+RF_LIGHTRES) * sw) / 100.0f) && y >= (sh - (int)(((bottom+RF_LIGHTRES) * sh) / 100.0f)) && y <= (sh - (int)((bottom * sh) / 100.0f)))
			{
				if(selectedLightQty>0)
				{
				float pixelSizeX = (float)sw / 100.0f;
				float pixelSizeY = (float)sh / 100.0f;

				int xIndex = (int)(((float)x / pixelSizeX)-left);
				int yIndex = (int)(((float)(sh-y) / pixelSizeY)-bottom);
				if(xIndex != (int)myLights[selectedLight].xIndex || yIndex != (int)myLights[selectedLight].yIndex)
				{
					subLight(selectedLight);
					myLights[selectedLight].xIndex = xIndex;
					myLights[selectedLight].yIndex = yIndex;
					addLight(selectedLight);
					bChange = true;
					glutPostRedisplay();
				}
				moving = 1;
				}
				return;
			}
		}

		if (button == GLUT_LEFT_BUTTON && state == GLUT_UP)
		{
			moving = 0;
			changingcolor = -1;
			return;
		}
	}

	// Check for button pressed to rotate environment map
	if(texFlag == RF_TEX_CUBEMAP || texFlag == RF_TEX_POINTCUBE)
	{
		if (button == GLUT_LEFT_BUTTON && state == GLUT_DOWN)
		{
			moving = 2;
			beginx = x;
			beginy = y;
		}

		if (button == GLUT_LEFT_BUTTON && state == GLUT_UP)
		{
			moving = 0;
		}
	}

}
// =========================================================================================


// =========================================================================================
//  Idle
// -----------------------------------------------------------------------------------------
void idle()
{
	// To generate video file (rotation matrices) if flag active
	if(saveMatrices)
	{
		currentTime = clock();
		elapsedTime = (currentTime - lastTime);
		if(elapsedTime >= RF_FREQUENCY)
		{
			lastTime = currentTime - (elapsedTime - RF_FREQUENCY);
			genVideo();
		}
	}

	// To generate screenshots if flag active
	if(saveScreenshot)
	{
		currentTime = clock();
		elapsedTime = (currentTime - lastTime);
		if(elapsedTime >= RF_FREQUENCY)
		{
			lastTime = currentTime - (elapsedTime - RF_FREQUENCY);
			char myScreenshot[1024];
			sprintf(myScreenshot, "%s%.8d.tga", RF_VIDEODIR, filenumber);
			screenshot(myScreenshot);
			printf("--> File %s saved.\n",myScreenshot);
			filenumber++;
			bChange = true;
		}
	}

	// To recompute image if loading rotation matrices to save images
	if(loadMatrices)
	{
		bChange = true;
		glutPostRedisplay();
	}

	if (changingcolor!=-1 && texFlag == RF_TEX_NOTEXTURE || texFlag == RF_TEX_POINTCUBE)
	{
		switch(changingcolor)
		{
		case 0: keyboardSpecial(GLUT_KEY_INSERT, 0, 0); break;
		case 1: keyboard(127, 0, 0); break;
		case 2: keyboardSpecial(GLUT_KEY_HOME, 0, 0); break;
		case 3: keyboardSpecial(GLUT_KEY_END, 0, 0); break;
		case 4: keyboardSpecial(GLUT_KEY_PAGE_UP, 0, 0); break;
		case 5: keyboardSpecial(GLUT_KEY_PAGE_DOWN, 0, 0); break;
		}
	}
}
// =========================================================================================


// =========================================================================================
//  Motion
// -----------------------------------------------------------------------------------------
void motion(int x, int y)
{
	// To generate video file (rotation matrices) if flag active
	if(saveMatrices)
	{
		currentTime = clock();
		elapsedTime = (currentTime - lastTime);
		if(elapsedTime >= RF_FREQUENCY)
		{
			lastTime = currentTime - (elapsedTime - RF_FREQUENCY);
			genVideo();
		}
	}

	// To generate screenshots if flag active
	if(saveScreenshot)
	{
		currentTime = clock();
		elapsedTime = (currentTime - lastTime);
		if(elapsedTime >= RF_FREQUENCY)
		{
			lastTime = currentTime - (elapsedTime - RF_FREQUENCY);
			char myScreenshot[1024];
			sprintf(myScreenshot, "%s%.8d.tga", RF_VIDEODIR, filenumber);
			screenshot(myScreenshot);
			printf("--> File %s saved.\n",myScreenshot);
			filenumber++;
			bChange = true;
		}
	}

	// Track mouse motion to rotate environment map
	if(texFlag == RF_TEX_CUBEMAP || texFlag == RF_TEX_POINTCUBE)
	{
		if (moving==2)
		{
			trackball(lastquat,
				(2.0 * beginx - sw) / sw,
				(sh - 2.0 * beginy) / sh,
				(2.0 * x - sw) / sw,
				(sh - 2.0 * y) / sh		);
			beginx = x;
			beginy = y;
			add_quats(lastquat, curquat, curquat);
			bChange = true;
			glutPostRedisplay();
		}
	}

	// Track mouse motion to move point lights
	int left = (RF_LIGHTSPANEL_X0 - RF_LIGHTRES - 2);
	int bottom = (RF_LIGHTSPANEL_Y0 - RF_LIGHTRES);
	if(texFlag == RF_TEX_NOTEXTURE || texFlag == RF_TEX_POINTCUBE)
	{
		if (moving==1)
		{
			if(x >= (int)((left * sw) / 100.0f) && x <= (int)(((left+RF_LIGHTRES) * sw) / 100.0f) && y >= (sh - (int)(((bottom+RF_LIGHTRES) * sh) / 100.0f)) && y <= (sh - (int)((bottom * sh) / 100.0f)))
			{
				float pixelSizeX = (float)sw / 100.0f;
				float pixelSizeY = (float)sh / 100.0f;

				int xIndex = (int)(((float)x / pixelSizeX)-left);
				int yIndex = (int)(((float)(sh-y) / pixelSizeY)-bottom);

				if(xIndex != (int)myLights[selectedLight].xIndex || yIndex != (int)myLights[selectedLight].yIndex)
				{
					subLight(selectedLight);
					myLights[selectedLight].xIndex = xIndex;
					myLights[selectedLight].yIndex = yIndex;
					addLight(selectedLight);
					bChange = true;
					glutPostRedisplay();
				}
			}
		}

		if (changingcolor!=-1)
		{
			switch(changingcolor)
			{
			case 0: keyboardSpecial(GLUT_KEY_INSERT, x, y); break;
			case 1: keyboard(127, x, y); break;
			case 2: keyboardSpecial(GLUT_KEY_HOME, x, y); break;
			case 3: keyboardSpecial(GLUT_KEY_END, x, y); break;
			case 4: keyboardSpecial(GLUT_KEY_PAGE_UP, x, y); break;
			case 5: keyboardSpecial(GLUT_KEY_PAGE_DOWN, x, y); break;
			}
		}
	}

}
// =========================================================================================


// =========================================================================================
//  Main
// -----------------------------------------------------------------------------------------
int main(int argc, char **argv)
{
	glutInitWindowSize(SCREEN_WIDTH, SCREEN_HEIGHT);
	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_DOUBLE | GLUT_DEPTH | GLUT_RGB);
	glutCreateWindow("Relighting Framework :: Columbia University :: January 2004 :: Sebastian Enrique [senrique@cs.columbia.edu]");

	glutDisplayFunc(display);
	glutReshapeFunc(reshape);
	glutKeyboardFunc(keyboard);
	glutSpecialFunc(keyboardSpecial);
	glutMouseFunc(mouse);
	glutMotionFunc(motion);
	glutIdleFunc(idle);

	glutCreateMenu(menu);
	glutAddMenuEntry("Point Lights", RF_TEX_NOTEXTURE);
	glutAddMenuEntry("Cube Map", RF_TEX_CUBEMAP);
	glutAddMenuEntry("Point Lights + Cube Map", RF_TEX_POINTCUBE);
	glutAttachMenu(GLUT_RIGHT_BUTTON);

	glViewport (0, 0, SCREEN_WIDTH, SCREEN_HEIGHT);				// Reset The Current Viewport
	glMatrixMode (GL_PROJECTION);								// Select The Projection Matrix
	glLoadIdentity ();											// Reset The Projection Matrix
	gluOrtho2D(0.0f, 100.0f, 0.0, 100.0f);						// Ortographic Projection
	glMatrixMode (GL_MODELVIEW);								// Select The Modelview Matrix
	glLoadIdentity ();											// Reset The Modelview Matrix

	glClearColor (0.0f, 0.0f, 0.0f, 0.5f);						// Black Background
	glClearDepth (1.0f);										// Depth Buffer Setup
	glDepthFunc (GL_LEQUAL);									// The Type Of Depth Testing (Less Or Equal)
	glEnable (GL_DEPTH_TEST);									// Enable Depth Testing
	glShadeModel (GL_SMOOTH);									// Select Smooth Shading

	moving=0;

	if(argc!=2)	{ printf("You need to specify the control file: rf <control.rfc>\n"); return 1; }
	if(RFC_parse(argv[1])!=0) return 1;

	MyInit();

	glutMainLoop();

	if(myVideoFile!=NULL) fclose(myVideoFile);

	return 0;
}
// =========================================================================================

// ***************************************************************************************************
// ***************************************************************************************************
