// ***************************************************************************************************
// ***************************************************************************************************
//  RELIGHTING FRAMEWORK v0.10
//  Sebastian Enrique - senrique@cs.columbia.edu
//  Columbia University
//  January 2004
// ***************************************************************************************************
// ***************************************************************************************************

/*
	RELIGHTING FRAMEWORK CONTROL FILE SPECIFICATION
	
	  Name: xxxx.rfc		(Relighting Framework Control)
      Conditions:	Each cube face should not be greater than 256x256.

	Implemented Tokens:

	IMAGERESW = 512				// INT Image to relight is 512 pixels width
	IMAGERESH = 512				// INT Image to relight is 512 pixels height
	LIGHTRES = 32				// INT Each cube face has 32x32 point light positions
	BLOCKRESW = 32				// INT The image is divided in 32 blocks in width
	BLOCKRESH = 32				// INT The image is divided in 32 blocks in height
	PIXELRES = 16				// INT Each block contains 16x16 pixels. Note that IMAGERES = BLOCKRES * PIXELRES
	RANKB = 10					// INT Rank applied to the first SVD computation
	RANKQ = 20					// INT Rank applied to the second SVD computation
	HDRFILE = kitchen_cross.hdr	// STR Filename and path for the high dynamic range file to use as cubemap
	ERFILE = er.txt				// STR Filename and path for red color component E matrix
	EGFILE = eg.txt				// STR Filename and path for red color component E matrix
	EBFILE = eb.txt				// STR Filename and path for red color component E matrix
	VRFILE = vr.txt				// STR Filename and path for green color component V matrix
	VGFILE = vg.txt				// STR Filename and path for green color component V matrix
	VBFILE = vb.txt				// STR Filename and path for green color component V matrix
	URFILE = ur.txt				// STR Filename and path for blue color component U matrix
	UGFILE = ug.txt				// STR Filename and path for blue color component U matrix
	UBFILE = ub.txt				// STR Filename and path for blue color component U matrix

	One token per line. Tokens can be in any order.


	Tokens Wish List:

	SCREEN_WIDTH = 1024		// INT Starting application screen width
	SCREEN_HEIGHT = 768		// INT Starting application screen height

	DELTA_COLOR = 0.05		// FP Step for incrementing/decrementing a point light color component
  
	CROSS_XC = 85.0			// FP X center position to display the cubemap cross 0.0 <= CROSS_XC <= 100.0
	CROSS_YC = 20.0			// FP Y center position to display the cubemap cross 0.0 <= CROSS_YC <= 100.0 

	LIGHTSPANEL_X0 = 33.0	// FP Left position to display the light information panel 0.0 <= LIGHTSPANEL_X0 <= 100.0
	LIGHTSPANEL_Y0 = 18.5	// FP Top position to display the light information panel 0.0 <= LIGHTSPANEL_Y0 <= 100.0
	LIGHTSPANEL_SCALE = 1.5	// FP Scale to apply to light information panel 0.0 <= LIGHTSPANEL_X0 <= 100.0
	LIGHTSPANEL_YSTEP = 2.0	// FP Space between each light information

*/

#include <stdio.h>
#include <string.h>
#include <windows.h>

// =================================================================================================================
//  TOKEN DEFINITIONS
// -----------------------------------------------------------------------------------------------------------------
#define TOKEN_IMAGERESW	0x00001
#define TOKEN_LIGHTRES	0x00002
#define TOKEN_BLOCKRESW	0x00004
#define TOKEN_PIXELRES	0x00008
#define TOKEN_RANKB		0x00010
#define TOKEN_RANKQ		0x00020
#define TOKEN_HDRFILE	0x00040
#define TOKEN_ERFILE	0x00080
#define TOKEN_EGFILE	0x00100
#define TOKEN_EBFILE	0x00200
#define TOKEN_VRFILE	0x00400
#define TOKEN_VGFILE	0x00800
#define TOKEN_VBFILE	0x01000
#define TOKEN_URFILE	0x02000
#define TOKEN_UGFILE	0x04000
#define TOKEN_UBFILE	0x08000
#define TOKEN_IMAGERESH	0x10000
#define TOKEN_BLOCKRESH 0x20000

#define TOKEN_ALL		0x3FFFF
#define TOKEN_NONE		0x00000
// =================================================================================================================

// =================================================================================================================
//  ERROR ENUMERATION
// -----------------------------------------------------------------------------------------------------------------
enum {
	RFC_OK = 0,
	RFC_ERR_FILENOTFOUND,
	RFC_ERR_TOKENDUPLICATED,
	RFC_ERR_INVALIDVALUE,
	RFC_ERR_INCOMPLETETOKENS,
	RFC_ERR_UNKNOWNTOKEN,
};
// =================================================================================================================


// =================================================================================================================
//  STRUCT WITH INCOMING DATA
// -----------------------------------------------------------------------------------------------------------------
struct RFCData {
	unsigned int imageresw;
	unsigned int imageresh;
	unsigned int lightres;
	unsigned int blockresw;
	unsigned int blockresh;
	unsigned int pixelres;
	unsigned int rankb;
	unsigned int rankq;
	char hdrfile[1024];
	char erfile[1024];
	char egfile[1024];
	char ebfile[1024];
	char vrfile[1024];
	char vgfile[1024];
	char vbfile[1024];
	char urfile[1024];
	char ugfile[1024];
	char ubfile[1024];
} myRFCData;
// =================================================================================================================


// =================================================================================================================
//  Main Function: Parse a rfc file
// -----------------------------------------------------------------------------------------------------------------
int RFC_parse(const char* myFile)
{
	FILE* myRFCFile;
	char str[2048];
	char filename[1024];
	int value;
	unsigned long flag;
	unsigned char found;
	unsigned int line;

	myRFCFile = fopen(myFile, "rb");

	if(myRFCFile == NULL)
	{
		printf("PARSING ERROR: Relighting Control File %s not found.\n",filename);
		return RFC_ERR_FILENOTFOUND;
	}

	flag = TOKEN_NONE;
	line = 0;
	for(;;)
	{
		line++;
		found = 0;
		if(fgets(str, sizeof(str)/sizeof(str[0]), myRFCFile)==NULL) break;

		if(sscanf(str," %s ", filename)==NULL)
			break;
		if(filename[0]!=0)
		{

		if(sscanf(str," IMAGERESW = %d ", &value)!=NULL)
		{
			if((flag & TOKEN_IMAGERESW)!=0)
			{
				fclose(myRFCFile);
				printf("PARSING ERROR: Token IMAGERESW duplicated, line %d.\n",line);
				return RFC_ERR_TOKENDUPLICATED;
			}
			if(value <= 0)
			{
				fclose(myRFCFile);
				printf("PARSING ERROR: IMAGERESW should be greater than 0, line %d.\n",line);
				return RFC_ERR_INVALIDVALUE;
			}
			myRFCData.imageresw = value;
			flag|=TOKEN_IMAGERESW;
			if((flag & (TOKEN_IMAGERESW|TOKEN_BLOCKRESW|TOKEN_PIXELRES))==(TOKEN_IMAGERESW|TOKEN_BLOCKRESW|TOKEN_PIXELRES))
			{
				if(myRFCData.imageresw != (myRFCData.blockresw * myRFCData.pixelres))
				{
					fclose(myRFCFile);
					printf("PARSING ERROR: IMAGERESW != BLOCKRESW * PIXELRES, line %d.\n",line);
					return RFC_ERR_INVALIDVALUE;
				}
			}
			found = 1;
		}

		if(sscanf(str," IMAGERESH = %d ", &value)!=NULL)
		{
			if((flag & TOKEN_IMAGERESH)!=0)
			{
				fclose(myRFCFile);
				printf("PARSING ERROR: Token IMAGERESH duplicated, line %d.\n",line);
				return RFC_ERR_TOKENDUPLICATED;
			}
			if(value <= 0)
			{
				fclose(myRFCFile);
				printf("PARSING ERROR: IMAGERESH should be greater than 0, line %d.\n",line);
				return RFC_ERR_INVALIDVALUE;
			}
			myRFCData.imageresh = value;
			flag|=TOKEN_IMAGERESH;
			if((flag & (TOKEN_IMAGERESH|TOKEN_BLOCKRESH|TOKEN_PIXELRES))==(TOKEN_IMAGERESH|TOKEN_BLOCKRESH|TOKEN_PIXELRES))
			{
				if(myRFCData.imageresh != (myRFCData.blockresh * myRFCData.pixelres))
				{
					fclose(myRFCFile);
					printf("PARSING ERROR: IMAGERESH != BLOCKRESH * PIXELRES, line %d.\n",line);
					return RFC_ERR_INVALIDVALUE;
				}
			}
			found = 1;
		}

		if(sscanf(str," LIGHTRES = %d ", &value)!=NULL)
		{
			if((flag & TOKEN_LIGHTRES)!=0)
			{
				fclose(myRFCFile);
				printf("PARSING ERROR: Token LIGHTRES duplicated, line %d.\n",line);
				return RFC_ERR_TOKENDUPLICATED;
			}
			if(value <= 0)
			{
				fclose(myRFCFile);
				printf("PARSING ERROR: LIGHTRES should be greater than 0, line %d.\n",line);
				return RFC_ERR_INVALIDVALUE;
			}
			myRFCData.lightres = value;
			flag|=TOKEN_LIGHTRES;
			found = 1;
		}

		if(sscanf(str," BLOCKRESW = %d ", &value)!=NULL)
		{
			if((flag & TOKEN_BLOCKRESW)!=0)
			{
				fclose(myRFCFile);
				printf("PARSING ERROR: Token BLOCKRESW duplicated, line %d.\n",line);
				return RFC_ERR_TOKENDUPLICATED;
			}
			if(value <= 0)
			{
				fclose(myRFCFile);
				printf("PARSING ERROR: BLOCKRESW should be greater than 0, line %d.\n",line);
				return RFC_ERR_INVALIDVALUE;
			}
			myRFCData.blockresw = value;
			flag|=TOKEN_BLOCKRESW;
			if((flag & (TOKEN_IMAGERESW|TOKEN_BLOCKRESW|TOKEN_PIXELRES))==(TOKEN_IMAGERESW|TOKEN_BLOCKRESW|TOKEN_PIXELRES))
			{
				if(myRFCData.imageresw != (myRFCData.blockresw * myRFCData.pixelres))
				{
					fclose(myRFCFile);
					printf("PARSING ERROR: IMAGERESW != BLOCKRESW * PIXELRES, line %d.\n",line);
					return RFC_ERR_INVALIDVALUE;
				}
			}
			found = 1;
		}

		if(sscanf(str," BLOCKRESH = %d ", &value)!=NULL)
		{
			if((flag & TOKEN_BLOCKRESH)!=0)
			{
				fclose(myRFCFile);
				printf("PARSING ERROR: Token BLOCKRES duplicated, line %d.\n",line);
				return RFC_ERR_TOKENDUPLICATED;
			}
			if(value <= 0)
			{
				fclose(myRFCFile);
				printf("PARSING ERROR: BLOCKRESH should be greater than 0, line %d.\n",line);
				return RFC_ERR_INVALIDVALUE;
			}
			myRFCData.blockresh = value;
			flag|=TOKEN_BLOCKRESH;
			if((flag & (TOKEN_IMAGERESH|TOKEN_BLOCKRESH|TOKEN_PIXELRES))==(TOKEN_IMAGERESH|TOKEN_BLOCKRESH|TOKEN_PIXELRES))
			{
				if(myRFCData.imageresh != (myRFCData.blockresh * myRFCData.pixelres))
				{
					fclose(myRFCFile);
					printf("PARSING ERROR: IMAGERESH != BLOCKRESH * PIXELRES, line %d.\n",line);
					return RFC_ERR_INVALIDVALUE;
				}
			}
			found = 1;
		}

		if(sscanf(str," PIXELRES = %d ", &value)!=NULL)
		{
			if((flag & TOKEN_PIXELRES)!=0)
			{
				fclose(myRFCFile);
				printf("PARSING ERROR: Token PIXELRES duplicated, line %d.\n",line);
				return RFC_ERR_TOKENDUPLICATED;
			}
			if(value <= 0)
			{
				fclose(myRFCFile);
				printf("PARSING ERROR: PIXELRES should be greater than 0, line %d.\n",line);
				return RFC_ERR_INVALIDVALUE;
			}
			myRFCData.pixelres = value;
			flag|=TOKEN_PIXELRES;
			if((flag & (TOKEN_IMAGERESW|TOKEN_BLOCKRESW|TOKEN_PIXELRES))==(TOKEN_IMAGERESW|TOKEN_BLOCKRESW|TOKEN_PIXELRES))
			{
				if(myRFCData.imageresw != (myRFCData.blockresw * myRFCData.pixelres))
				{
					fclose(myRFCFile);
					printf("PARSING ERROR: IMAGERESW != BLOCKRESW * PIXELRES, line %d.\n",line);
					return RFC_ERR_INVALIDVALUE;
				}
			}
			if((flag & (TOKEN_IMAGERESH|TOKEN_BLOCKRESH|TOKEN_PIXELRES))==(TOKEN_IMAGERESH|TOKEN_BLOCKRESH|TOKEN_PIXELRES))
			{
				if(myRFCData.imageresh != (myRFCData.blockresh * myRFCData.pixelres))
				{
					fclose(myRFCFile);
					printf("PARSING ERROR: IMAGERESH != BLOCKRESH * PIXELRES, line %d.\n",line);
					return RFC_ERR_INVALIDVALUE;
				}
			}
			found = 1;
		}

		if(sscanf(str," RANKB = %d ", &value)!=NULL)
		{
			if((flag & TOKEN_RANKB)!=0)
			{
				fclose(myRFCFile);
				printf("PARSING ERROR: Token RANKB duplicated, line %d.\n",line);
				return RFC_ERR_TOKENDUPLICATED;
			}
			if(value <= 0)
			{
				fclose(myRFCFile);
				printf("PARSING ERROR: RANKB should be greater than 0, line %d.\n",line);
				return RFC_ERR_INVALIDVALUE;
			}
			myRFCData.rankb = value;
			flag|=TOKEN_RANKB;
			found = 1;
		}

		if(sscanf(str," RANKQ = %d ", &value)!=NULL)
		{
			if((flag & TOKEN_RANKQ)!=0)
			{
				fclose(myRFCFile);
				printf("PARSING ERROR: Token RANKQ duplicated, line %d.\n",line);
				return RFC_ERR_TOKENDUPLICATED;
			}
			if(value <= 0)
			{
				fclose(myRFCFile);
				printf("PARSING ERROR: RANKQ should be greater than 0, line %d.\n",line);
				return RFC_ERR_INVALIDVALUE;
			}
			myRFCData.rankq = value;
			flag|=TOKEN_RANKQ;
			found = 1;
		}

		if(sscanf(str," HDRFILE = %s ", filename)!=NULL)
		{
			if((flag & TOKEN_HDRFILE)!=0)
			{
				fclose(myRFCFile);
				printf("PARSING ERROR: Token HDRFILE duplicated, line %d.\n",line);
				return RFC_ERR_TOKENDUPLICATED;
			}
			strcpy(myRFCData.hdrfile, filename);
			flag|=TOKEN_HDRFILE;
			found = 1;
		}

		if(sscanf(str," ERFILE = %s ", filename)!=NULL)
		{
			if((flag & TOKEN_ERFILE)!=0)
			{
				fclose(myRFCFile);
				printf("PARSING ERROR: Token ERFILE duplicated, line %d.\n",line);
				return RFC_ERR_TOKENDUPLICATED;
			}
			strcpy(myRFCData.erfile, filename);
			flag|=TOKEN_ERFILE;
			found = 1;
		}

		if(sscanf(str," EGFILE = %s ", filename)!=NULL)
		{
			if((flag & TOKEN_EGFILE)!=0)
			{
				fclose(myRFCFile);
				printf("PARSING ERROR: Token EGFILE duplicated, line %d.\n",line);
				return RFC_ERR_TOKENDUPLICATED;
			}
			strcpy(myRFCData.egfile, filename);
			flag|=TOKEN_EGFILE;
			found = 1;
		}

		if(sscanf(str," EBFILE = %s ", filename)!=NULL)
		{
			if((flag & TOKEN_EBFILE)!=0)
			{
				fclose(myRFCFile);
				printf("PARSING ERROR: Token EBFILE duplicated, line %d.\n",line);
				return RFC_ERR_TOKENDUPLICATED;
			}
			strcpy(myRFCData.ebfile, filename);
			flag|=TOKEN_EBFILE;
			found = 1;
		}

		if(sscanf(str," VRFILE = %s ", filename)!=NULL)
		{
			if((flag & TOKEN_VRFILE)!=0)
			{
				fclose(myRFCFile);
				printf("PARSING ERROR: Token VRFILE duplicated, line %d.\n",line);
				return RFC_ERR_TOKENDUPLICATED;
			}
			strcpy(myRFCData.vrfile, filename);
			flag|=TOKEN_VRFILE;
			found = 1;
		}

		if(sscanf(str," VGFILE = %s ", filename)!=NULL)
		{
			if((flag & TOKEN_VGFILE)!=0)
			{
				fclose(myRFCFile);
				printf("PARSING ERROR: Token VGFILE duplicated, line %d.\n",line);
				return RFC_ERR_TOKENDUPLICATED;
			}
			strcpy(myRFCData.vgfile, filename);
			flag|=TOKEN_VGFILE;
			found = 1;
		}

		if(sscanf(str," VBFILE = %s ", filename)!=NULL)
		{
			if((flag & TOKEN_VBFILE)!=0)
			{
				fclose(myRFCFile);
				printf("PARSING ERROR: Token VBFILE duplicated, line %d.\n",line);
				return RFC_ERR_TOKENDUPLICATED;
			}
			strcpy(myRFCData.vbfile, filename);
			flag|=TOKEN_VBFILE;
			found = 1;
		}

		if(sscanf(str," URFILE = %s ", filename)!=NULL)
		{
			if((flag & TOKEN_URFILE)!=0)
			{
				fclose(myRFCFile);
				printf("PARSING ERROR: Token URFILE duplicated, line %d.\n",line);
				return RFC_ERR_TOKENDUPLICATED;
			}
			strcpy(myRFCData.urfile, filename);
			flag|=TOKEN_URFILE;
			found = 1;
		}

		if(sscanf(str," UGFILE = %s ", filename)!=NULL)
		{
			if((flag & TOKEN_UGFILE)!=0)
			{
				fclose(myRFCFile);
				printf("PARSING ERROR: Token UGFILE duplicated, line %d.\n",line);
				return RFC_ERR_TOKENDUPLICATED;
			}
			strcpy(myRFCData.ugfile, filename);
			flag|=TOKEN_UGFILE;
			found = 1;
		}

		if(sscanf(str," UBFILE = %s ", filename)!=NULL)
		{
			if((flag & TOKEN_UBFILE)!=0)
			{
				fclose(myRFCFile);
				printf("PARSING ERROR: Token UBFILE duplicated, line %d.\n",line);
				return RFC_ERR_TOKENDUPLICATED;
			}
			strcpy(myRFCData.ubfile, filename);
			flag|=TOKEN_UBFILE;
			found = 1;
		}

		if(found==0)
		{
			if(sscanf(str," %s ", filename)==NULL)
				break;
			if(filename[0]!=0)
			{
				fclose(myRFCFile);
				printf("PARSING ERROR: Unknown token, line %d.\n",line);
				return RFC_ERR_UNKNOWNTOKEN;
			}
		}
		}
	}

	fclose(myRFCFile);

	if((flag & TOKEN_ALL)!= TOKEN_ALL)
	{
		// Not all the tokens were specified
		if((flag & TOKEN_IMAGERESW)==0) printf("PARSING ERROR: Unable to find token IMAGERESW.\n");
		if((flag & TOKEN_IMAGERESH)==0) printf("PARSING ERROR: Unable to find token IMAGERESH.\n");
		if((flag & TOKEN_LIGHTRES)==0) printf("PARSING ERROR: Unable to find token LIGHTRES.\n");
		if((flag & TOKEN_BLOCKRESW)==0) printf("PARSING ERROR: Unable to find token BLOCKRESW.\n");
		if((flag & TOKEN_BLOCKRESH)==0) printf("PARSING ERROR: Unable to find token BLOCKRESH.\n");
		if((flag & TOKEN_PIXELRES)==0) printf("PARSING ERROR: Unable to find token PIXELRES.\n");
		if((flag & TOKEN_RANKB)==0) printf("PARSING ERROR: Unable to find token RANKB.\n");
		if((flag & TOKEN_RANKQ)==0) printf("PARSING ERROR: Unable to find token RANKQ.\n");
		if((flag & TOKEN_HDRFILE)==0) printf("PARSING ERROR: Unable to find token HDRFILE.\n");
		if((flag & TOKEN_ERFILE)==0) printf("PARSING ERROR: Unable to find token ERFILE.\n");
		if((flag & TOKEN_EGFILE)==0) printf("PARSING ERROR: Unable to find token EGFILE.\n");
		if((flag & TOKEN_EBFILE)==0) printf("PARSING ERROR: Unable to find token EBFILE.\n");
		if((flag & TOKEN_VRFILE)==0) printf("PARSING ERROR: Unable to find token VRFILE.\n");
		if((flag & TOKEN_VGFILE)==0) printf("PARSING ERROR: Unable to find token BGFILE.\n");
		if((flag & TOKEN_VBFILE)==0) printf("PARSING ERROR: Unable to find token VBFILE.\n");
		if((flag & TOKEN_URFILE)==0) printf("PARSING ERROR: Unable to find token URFILE.\n");
		if((flag & TOKEN_UGFILE)==0) printf("PARSING ERROR: Unable to find token UGFILE.\n");
		if((flag & TOKEN_UBFILE)==0) printf("PARSING ERROR: Unable to find token UBFILE.\n");

		return RFC_ERR_INCOMPLETETOKENS;
	}

	return RFC_OK;
}
// =================================================================================================================

// ***************************************************************************************************
// ***************************************************************************************************
