/*
	RELIGHTING FRAMEWORK CONTROL FILE SPECIFICATION
	
	  Name: xxxx.rfc		(Relighting Framework Control)
      Conditions:	Each cube face should not be greater than 256x256.

	Implemented Tokens:

	IMAGERES = 512				// INT Image to relight is 512x512
	LIGHTRES = 32				// INT Each cube face has 32x32 point light positions
	BLOCKRES = 32				// INT The image is divided in 32x32 blocks
	PIXELRES = 16				// INT Each block contains 16x16 pixels. Note that IMAGERES = BLOCKRES * PIXELRES
	RANKB = 10					// INT Rank applied to the first SVD computation
	RANKQ = 20					// INT Rank applied to the second SVD computation
	HDRFILE = kitchen_cross.hdr	// STR Filename and path for the high dynamic range file to use as cubemap
	ERFILE = er.txt				// STR Filename and path for red color component E matrix
	EGFILE = eg.txt				// STR Filename and path for red color component E matrix
	EBFILE = eb.txt				// STR Filename and path for red color component E matrix
	VRFILE = vr.txt				// STR Filename and path for green color component V matrix
	VGFILE = vg.txt				// STR Filename and path for green color component V matrix
	VBFILE = vb.txt				// STR Filename and path for green color component V matrix
	URFILE = ur.txt				// STR Filename and path for blue color component U matrix
	UGFILE = ug.txt				// STR Filename and path for blue color component U matrix
	UBFILE = ub.txt				// STR Filename and path for blue color component U matrix

	One token per line. Tokens can be in any order.


	Tokens Wish List:

	SCREEN_WIDTH = 1024		// INT Starting application screen width
	SCREEN_HEIGHT = 768		// INT Starting application screen height

	DELTA_COLOR = 0.05		// FP Step for incrementing/decrementing a point light color component
  
	CROSS_XC = 85.0			// FP X center position to display the cubemap cross 0.0 <= CROSS_XC <= 100.0
	CROSS_YC = 20.0			// FP Y center position to display the cubemap cross 0.0 <= CROSS_YC <= 100.0 

	LIGHTSPANEL_X0 = 33.0	// FP Left position to display the light information panel 0.0 <= LIGHTSPANEL_X0 <= 100.0
	LIGHTSPANEL_Y0 = 18.5	// FP Top position to display the light information panel 0.0 <= LIGHTSPANEL_Y0 <= 100.0
	LIGHTSPANEL_SCALE = 1.5	// FP Scale to apply to light information panel 0.0 <= LIGHTSPANEL_X0 <= 100.0
	LIGHTSPANEL_YSTEP = 2.0	// FP Space between each light information

*/

#include "rfcparser.h"
#include <stdio.h>
#include <string.h>

#define TOKEN_IMAGERES	0x0001
#define TOKEN_LIGHTRES	0x0002
#define TOKEN_BLOCKRES	0x0004
#define TOKEN_PIXELRES	0x0008
#define TOKEN_RANKB		0x0010
#define TOKEN_RANKQ		0x0020
#define TOKEN_HDRFILE	0x0040
#define TOKEN_ERFILE	0x0080
#define TOKEN_EGFILE	0x0100
#define TOKEN_EBFILE	0x0200
#define TOKEN_VRFILE	0x0400
#define TOKEN_VGFILE	0x0800
#define TOKEN_VBFILE	0x1000
#define TOKEN_URFILE	0x2000
#define TOKEN_UGFILE	0x4000
#define TOKEN_UBFILE	0x8000

#define TOKEN_ALL		0xFFFF
#define TOKEN_NONE		0x0000

enum {
	RFC_OK = 0,
	RFC_ERR_FILENOTFOUND,
	RFC_ERR_TOKENDUPLICATED,
	RFC_ERR_INVALIDVALUE,
	RFC_ERR_INCOMPLETETOKENS,
	RFC_ERR_UNKNOWNTOKEN,
};

int RFC_parse(const char* myFile)
{
	FILE* myRFCFile;
	char str[2048];
	char filename[1024];
	int value;
	unsigned int flag;
	unsigned char found;
	unsigned int line;

	myRFCFile = fopen(myFile, "rb");

	if(myRFCFile == NULL)
	{
		printf("PARSING ERROR: Relighting Control File %s not found.\n",filename);
		return RFC_ERR_FILENOTFOUND;
	}

	flag = TOKEN_NONE;
	line = 0;
	while(true)
	{
		line++;
		found = 0;
		if(fgets(str, sizeof(str)/sizeof(str[0]), myRFCFile)==NULL) break;

		int test = sscanf(str," %s ", filename);
		if(test!=-1)
		{

		if(sscanf(str," IMAGERES = %d ", &value)!=NULL)
		{
			if((flag & TOKEN_IMAGERES)!=0)
			{
				fclose(myRFCFile);
				printf("PARSING ERROR: Token IMAGERES duplicated, line %d.\n",line);
				return RFC_ERR_TOKENDUPLICATED;
			}
			if(value <= 0)
			{
				fclose(myRFCFile);
				printf("PARSING ERROR: IMAGERES should be greater than 0, line %d.\n",line);
				return RFC_ERR_INVALIDVALUE;
			}
			myRFCData.imageres = value;
			flag|=TOKEN_IMAGERES;
			if((flag & (TOKEN_IMAGERES|TOKEN_BLOCKRES|TOKEN_PIXELRES))==(TOKEN_IMAGERES|TOKEN_BLOCKRES|TOKEN_PIXELRES))
			{
				if(myRFCData.imageres != (myRFCData.blockres * myRFCData.pixelres))
				{
					fclose(myRFCFile);
					printf("PARSING ERROR: IMAGERES != BLOCKRES * PIXELRES, line %d.\n",line);
					return RFC_ERR_INVALIDVALUE;
				}
			}
			found = 1;
		}

		if(sscanf(str," LIGHTRES = %d ", &value)!=NULL)
		{
			if((flag & TOKEN_LIGHTRES)!=0)
			{
				fclose(myRFCFile);
				printf("PARSING ERROR: Token LIGHTRES duplicated, line %d.\n",line);
				return RFC_ERR_TOKENDUPLICATED;
			}
			if(value <= 0)
			{
				fclose(myRFCFile);
				printf("PARSING ERROR: LIGHTRES should be greater than 0, line %d.\n",line);
				return RFC_ERR_INVALIDVALUE;
			}
			myRFCData.lightres = value;
			flag|=TOKEN_LIGHTRES;
			found = 1;
		}

		if(sscanf(str," BLOCKRES = %d ", &value)!=NULL)
		{
			if((flag & TOKEN_BLOCKRES)!=0)
			{
				fclose(myRFCFile);
				printf("PARSING ERROR: Token BLOCKRES duplicated, line %d.\n",line);
				return RFC_ERR_TOKENDUPLICATED;
			}
			if(value <= 0)
			{
				fclose(myRFCFile);
				printf("PARSING ERROR: BLOCKRES should be greater than 0, line %d.\n",line);
				return RFC_ERR_INVALIDVALUE;
			}
			myRFCData.blockres = value;
			flag|=TOKEN_BLOCKRES;
			if((flag & (TOKEN_IMAGERES|TOKEN_BLOCKRES|TOKEN_PIXELRES))==(TOKEN_IMAGERES|TOKEN_BLOCKRES|TOKEN_PIXELRES))
			{
				if(myRFCData.imageres != (myRFCData.blockres * myRFCData.pixelres))
				{
					fclose(myRFCFile);
					printf("PARSING ERROR: IMAGERES != BLOCKRES * PIXELRES, line %d.\n",line);
					return RFC_ERR_INVALIDVALUE;
				}
			}
			found = 1;
		}

		if(sscanf(str," PIXELRES = %d ", &value)!=NULL)
		{
			if((flag & TOKEN_PIXELRES)!=0)
			{
				fclose(myRFCFile);
				printf("PARSING ERROR: Token PIXELRES duplicated, line %d.\n",line);
				return RFC_ERR_TOKENDUPLICATED;
			}
			if(value <= 0)
			{
				fclose(myRFCFile);
				printf("PARSING ERROR: PIXELRES should be greater than 0, line %d.\n",line);
				return RFC_ERR_INVALIDVALUE;
			}
			myRFCData.pixelres = value;
			flag|=TOKEN_PIXELRES;
			if((flag & (TOKEN_IMAGERES|TOKEN_BLOCKRES|TOKEN_PIXELRES))==(TOKEN_IMAGERES|TOKEN_BLOCKRES|TOKEN_PIXELRES))
			{
				if(myRFCData.imageres != (myRFCData.blockres * myRFCData.pixelres))
				{
					fclose(myRFCFile);
					printf("PARSING ERROR: IMAGERES != BLOCKRES * PIXELRES, line %d.\n",line);
					return RFC_ERR_INVALIDVALUE;
				}
			}
			found = 1;
		}

		if(sscanf(str," RANKB = %d ", &value)!=NULL)
		{
			if((flag & TOKEN_RANKB)!=0)
			{
				fclose(myRFCFile);
				printf("PARSING ERROR: Token RANKB duplicated, line %d.\n",line);
				return RFC_ERR_TOKENDUPLICATED;
			}
			if(value <= 0)
			{
				fclose(myRFCFile);
				printf("PARSING ERROR: RANKB should be greater than 0, line %d.\n",line);
				return RFC_ERR_INVALIDVALUE;
			}
			myRFCData.rankb = value;
			flag|=TOKEN_RANKB;
			found = 1;
		}

		if(sscanf(str," RANKQ = %d ", &value)!=NULL)
		{
			if((flag & TOKEN_RANKQ)!=0)
			{
				fclose(myRFCFile);
				printf("PARSING ERROR: Token RANKQ duplicated, line %d.\n",line);
				return RFC_ERR_TOKENDUPLICATED;
			}
			if(value <= 0)
			{
				fclose(myRFCFile);
				printf("PARSING ERROR: RANKQ should be greater than 0, line %d.\n",line);
				return RFC_ERR_INVALIDVALUE;
			}
			myRFCData.rankq = value;
			flag|=TOKEN_RANKQ;
			found = 1;
		}

		if(sscanf(str," HDRFILE = %s ", filename)!=NULL)
		{
			if((flag & TOKEN_HDRFILE)!=0)
			{
				fclose(myRFCFile);
				printf("PARSING ERROR: Token HDRFILE duplicated, line %d.\n",line);
				return RFC_ERR_TOKENDUPLICATED;
			}
			strcpy(myRFCData.hdrfile, filename);
			flag|=TOKEN_HDRFILE;
			found = 1;
		}

		if(sscanf(str," ERFILE = %s ", filename)!=NULL)
		{
			if((flag & TOKEN_ERFILE)!=0)
			{
				fclose(myRFCFile);
				printf("PARSING ERROR: Token ERFILE duplicated, line %d.\n",line);
				return RFC_ERR_TOKENDUPLICATED;
			}
			strcpy(myRFCData.erfile, filename);
			flag|=TOKEN_ERFILE;
			found = 1;
		}

		if(sscanf(str," EGFILE = %s ", filename)!=NULL)
		{
			if((flag & TOKEN_EGFILE)!=0)
			{
				fclose(myRFCFile);
				printf("PARSING ERROR: Token EGFILE duplicated, line %d.\n",line);
				return RFC_ERR_TOKENDUPLICATED;
			}
			strcpy(myRFCData.egfile, filename);
			flag|=TOKEN_EGFILE;
			found = 1;
		}

		if(sscanf(str," EBFILE = %s ", filename)!=NULL)
		{
			if((flag & TOKEN_EBFILE)!=0)
			{
				fclose(myRFCFile);
				printf("PARSING ERROR: Token EBFILE duplicated, line %d.\n",line);
				return RFC_ERR_TOKENDUPLICATED;
			}
			strcpy(myRFCData.ebfile, filename);
			flag|=TOKEN_EBFILE;
			found = 1;
		}

		if(sscanf(str," VRFILE = %s ", filename)!=NULL)
		{
			if((flag & TOKEN_VRFILE)!=0)
			{
				fclose(myRFCFile);
				printf("PARSING ERROR: Token VRFILE duplicated, line %d.\n",line);
				return RFC_ERR_TOKENDUPLICATED;
			}
			strcpy(myRFCData.vrfile, filename);
			flag|=TOKEN_VRFILE;
			found = 1;
		}

		if(sscanf(str," VGFILE = %s ", filename)!=NULL)
		{
			if((flag & TOKEN_VGFILE)!=0)
			{
				fclose(myRFCFile);
				printf("PARSING ERROR: Token VGFILE duplicated, line %d.\n",line);
				return RFC_ERR_TOKENDUPLICATED;
			}
			strcpy(myRFCData.vgfile, filename);
			flag|=TOKEN_VGFILE;
			found = 1;
		}

		if(sscanf(str," VBFILE = %s ", filename)!=NULL)
		{
			if((flag & TOKEN_VBFILE)!=0)
			{
				fclose(myRFCFile);
				printf("PARSING ERROR: Token VBFILE duplicated, line %d.\n",line);
				return RFC_ERR_TOKENDUPLICATED;
			}
			strcpy(myRFCData.vbfile, filename);
			flag|=TOKEN_VBFILE;
			found = 1;
		}

		if(sscanf(str," URFILE = %s ", filename)!=NULL)
		{
			if((flag & TOKEN_URFILE)!=0)
			{
				fclose(myRFCFile);
				printf("PARSING ERROR: Token URFILE duplicated, line %d.\n",line);
				return RFC_ERR_TOKENDUPLICATED;
			}
			strcpy(myRFCData.urfile, filename);
			flag|=TOKEN_URFILE;
			found = 1;
		}

		if(sscanf(str," UGFILE = %s ", filename)!=NULL)
		{
			if((flag & TOKEN_UGFILE)!=0)
			{
				fclose(myRFCFile);
				printf("PARSING ERROR: Token UGFILE duplicated, line %d.\n",line);
				return RFC_ERR_TOKENDUPLICATED;
			}
			strcpy(myRFCData.ugfile, filename);
			flag|=TOKEN_UGFILE;
			found = 1;
		}

		if(sscanf(str," UBFILE = %s ", filename)!=NULL)
		{
			if((flag & TOKEN_UBFILE)!=0)
			{
				fclose(myRFCFile);
				printf("PARSING ERROR: Token UBFILE duplicated, line %d.\n",line);
				return RFC_ERR_TOKENDUPLICATED;
			}
			strcpy(myRFCData.ubfile, filename);
			flag|=TOKEN_UBFILE;
			found = 1;
		}

		if(found==0)
		{
			if(sscanf(str," %s ", filename)==NULL)
				break;
			if(filename[0]!=0)
			{
				fclose(myRFCFile);
				printf("PARSING ERROR: Unknown token [%s], line %d.\n",filename,line);
				return RFC_ERR_UNKNOWNTOKEN;
			}
		}

		}
	}

	fclose(myRFCFile);

	if((flag & TOKEN_ALL)!= TOKEN_ALL)
	{
		// Not all the tokens were specified
		if((flag & TOKEN_IMAGERES)==0) printf("PARSING ERROR: Unable to find token IMAGERES.\n");
		if((flag & TOKEN_LIGHTRES)==0) printf("PARSING ERROR: Unable to find token LIGHTRES.\n");
		if((flag & TOKEN_BLOCKRES)==0) printf("PARSING ERROR: Unable to find token BLOCKRES.\n");
		if((flag & TOKEN_PIXELRES)==0) printf("PARSING ERROR: Unable to find token PIXELRES.\n");
		if((flag & TOKEN_RANKB)==0) printf("PARSING ERROR: Unable to find token RANKB.\n");
		if((flag & TOKEN_RANKQ)==0) printf("PARSING ERROR: Unable to find token RANKQ.\n");
		if((flag & TOKEN_HDRFILE)==0) printf("PARSING ERROR: Unable to find token HDRFILE.\n");
		if((flag & TOKEN_ERFILE)==0) printf("PARSING ERROR: Unable to find token ERFILE.\n");
		if((flag & TOKEN_EGFILE)==0) printf("PARSING ERROR: Unable to find token EGFILE.\n");
		if((flag & TOKEN_EBFILE)==0) printf("PARSING ERROR: Unable to find token EBFILE.\n");
		if((flag & TOKEN_VRFILE)==0) printf("PARSING ERROR: Unable to find token VRFILE.\n");
		if((flag & TOKEN_VGFILE)==0) printf("PARSING ERROR: Unable to find token BGFILE.\n");
		if((flag & TOKEN_VBFILE)==0) printf("PARSING ERROR: Unable to find token VBFILE.\n");
		if((flag & TOKEN_URFILE)==0) printf("PARSING ERROR: Unable to find token URFILE.\n");
		if((flag & TOKEN_UGFILE)==0) printf("PARSING ERROR: Unable to find token UGFILE.\n");
		if((flag & TOKEN_UBFILE)==0) printf("PARSING ERROR: Unable to find token UBFILE.\n");

		return RFC_ERR_INCOMPLETETOKENS;
	}

	return RFC_OK;
}

